"""A Twisted Python bridge for pqact exec'd processing"""

import hashlib
from datetime import datetime, timedelta, timezone
from io import BytesIO

from twisted.internet import stdio
from twisted.protocols import basic
from twisted.python import log


class LDMProductReceiver(basic.LineReceiver):
    """Our Protocol"""

    reactor = None
    product_start = b"\001"

    def __init__(self, dedup=False, isbinary=False, **kwargs):
        """Constructor

        Args:
          dedup (boolean): should we attempt to filter out duplicates
          isbinary (boolean): should we not attempt bytes decoding
            process_data will either return a string and bytes
          product_end (bytes): what is the end of a product from LDM, defaults
            to b'\r\r\n\003'
        """
        self.bytes_received = 0
        pe = kwargs.get("product_end")
        self.product_end = b"\r\r\n\003" if pe is None else pe
        self.productBuffer = BytesIO()
        # this puts twisted out of the pure line receiver mode
        self.setRawMode()
        self.cbFunc = self.process_data
        self.cache = {}
        self.isbinary = isbinary
        self.dedup = dedup
        if self.dedup:
            self.cbFunc = self.filter_product

    def clean_cache(self):
        """Cull old cache every 90 seconds"""
        threshold = datetime.now(timezone.utc) - timedelta(hours=4)
        # loop safety with this
        for digest in list(self.cache):
            if self.cache[digest] < threshold:
                self.cache.pop(digest)
        self.reactor.callLater(90, self.clean_cache)  # @UndefinedVariable

    def filter_product(self, original):
        """Implement Deduplication
         - Attempt to account for all of the wild variations that can happen
         with LDM plexing of NOAAPort, WeatherWire, TOC Socket Feed and
         whatever else may be happening

        1. If the character \x1e happens, we ignore it.  SPC issue here
        2. If we find \x17, this is some weather wire thing, we ignore whatever
           comes after it.
        3. We ignore any extraneous trailing space or line returns
        4. We convert tab characters to blank spaces, as one of NWSTG's systems
           does this already and is a source of duplicates
        5. Ignore first 11 bytes in MD5 computation

        If Okay, we end up calling self.process_data() with clean data

        Args:
          original (str): hopefully gaurenteed to be a string type
        """
        clean = original.replace("\x1e", "").replace("\t", "")
        if clean.find("\x17") > 0:
            clean = clean[: clean.find("\x17")]
        lines = clean.split("\015\015\012")
        # Trim trailing empty lines
        while lines and lines[-1].strip() == "":
            lines.pop()
        if len(lines) < 2:
            log.msg("ERROR: filter_product culled entire product (no data?)")
            return
        lines[1] = lines[1][:3]
        clean = "\015\015\012".join(lines)
        # first 11 characters should not be included in hex, like LDM does
        # hashlib works on bytes
        digest = hashlib.md5(
            clean[11:].encode("utf-8"), usedforsecurity=False
        ).hexdigest()
        if digest in self.cache:
            log.msg(f"DUP! {','.join(lines[1:5])}")
        else:
            self.cache[digest] = datetime.now(timezone.utc)
            self.process_data(clean + "\015\015\012")

    def rawDataReceived(self, data):
        """callback from twisted when raw data is received

        Args:
          data (str or bytes)
        """
        # write the data to our buffer
        self.bytes_received += self.productBuffer.write(data)

        # see how many products we may have
        tokens = self.productBuffer.getvalue().split(self.product_end)
        # If length tokens is 1, then we did not find the splitter
        if len(tokens) == 1:
            return

        # Everything up until the last one can always go...
        for token in tokens[:-1]:
            if self.isbinary:
                # we send bytes
                self.reactor.callLater(0, self.cbFunc, token)
            else:
                # we send str
                self.reactor.callLater(
                    0, self.cbFunc, token.decode("utf-8", "ignore")
                )
        # We have some cruft left over! be careful here as reassignment was
        # not working as expected, so we do things more cleanly
        self.productBuffer.seek(0)
        self.productBuffer.truncate()
        self.productBuffer.write(tokens[-1])

    def process_data(self, data):
        """callback function, either str or bytes depending on isbinary"""
        raise NotImplementedError

    def lineReceived(self, line):
        """needless override to make pylint happy"""
        log.msg(line)


class LDMProductFactory(stdio.StandardIO):
    """Our Factory"""

    def __init__(self, protocol, reactor=None, **kwargs):
        """constructor with a protocol instance"""
        if reactor is None:
            from twisted.internet import reactor
        protocol.reactor = reactor
        if protocol.dedup:
            reactor.callLater(90, protocol.clean_cache)
        stdio.StandardIO.__init__(self, protocol, reactor=reactor, **kwargs)



"""Make a file on disk appear more noaaport-ish."""

import sys

from pyiem.util import noaaport_text


def main(argv):
    """Do Main Things"""
    fn = argv[1]
    with open(fn, "rb") as fh:
        data = noaaport_text(fh.read().decode("ascii"))
    with open(fn, "w", encoding="ascii") as fh:
        fh.write(data)


if __name__ == "__main__":
    main(sys.argv)



"""Test the ldmbrige."""

from datetime import datetime, timezone
from unittest import mock

import pytest

from pywwa import ldmbridge


def test_process_data():
    """Test process_data."""
    proto = ldmbridge.LDMProductReceiver(dedup=True)
    with pytest.raises(NotImplementedError):
        proto.process_data("test\r\r\ntest\r\r\n")


def test_lineReceived():
    """Test lineReceived."""
    proto = ldmbridge.LDMProductReceiver(dedup=True)
    proto.reactor = mock.Mock()
    proto.process_data = mock.Mock()
    proto.cache["test"] = datetime.now(timezone.utc)
    proto.cache["test2"] = datetime(2007, 1, 1, 12, 14, tzinfo=timezone.utc)
    proto.clean_cache()
    proto.filter_product(r"test\r\r\ntest\r\r\ntest\x17\r\r\n\r\r\n")
    proto.filter_product(r"\r\r\n\r\r\n\r\r\n")
    proto.filter_product(r"test\r\r\ntest\r\r\ntest\x17\r\r\n\r\r\n")

    proto.rawDataReceived(
        b"\001000 \r\r\nSXUS50 KISU 010000\r\r\nADMA\r\r\n\003"
    )
    proto.lineReceived(b"000 \r\r\nSXUS50 KISU 010000\r\r\nADMA\r\r\n\003")



"""METAR product ingestor

NOTE: It is difficult to keep track of where I am getting the `Metar` library.
So let us document it here for my own sanity.

18 Jul 2017: `snowdepth` branch of my python-metar fork installed with pip
"""

# stdlib
from datetime import timezone

import click
from pyiem.nws.products import metarcollect
from pyiem.util import get_properties
from twisted.internet import reactor
from twisted.internet.task import LoopingCall

# Local
from pywwa import LOG, SETTINGS, common
from pywwa.database import get_database, load_metar_stations
from pywwa.ldm import bridge

IEMDB = get_database("iem")
ASOSDB = get_database("asos")
MESOSITEDB = get_database("mesosite")

NWSLI_PROVIDER = {}
# Manual list of sites that are sent to jabber :/
metarcollect.JABBER_SITES = {
    "KJFK": None,
    "KLGA": None,
    "KEWR": None,
    "KTEB": None,
    "KIAD": None,
    "KDCA": None,
    "KBWI": None,
    "KRIC": None,
    "KPHL": None,
}
# Sites with higher thresholds, hardcoded here for now :/
metarcollect.WIND_ALERT_THRESHOLD_KTS_BY_ICAO["KMWN"] = 80
# Try to prevent Jabber message dups
JABBER_MESSAGES = []
# List of sites to IGNORE and not send Jabber Messages for.
# iem property `pywwa_metar_ignorelist` should be a comma delimited 4 char ids
IGNORELIST = []


def load_ignorelist():
    """Sync what the properties database has for sites to ignore."""
    try:
        prop = get_properties().get("pywwa_metar_ignorelist", "")
        IGNORELIST.clear()
        for sid in [x.strip() for x in prop.split(",")]:
            if sid == "":
                continue
            if len(sid) != 4:
                LOG.msg("Not adding %s to IGNORELIST as not 4 char id", sid)
                continue
            IGNORELIST.append(sid)
    except Exception as exp:
        LOG.error(exp)
    LOG.info("Updated ignorelist is now %s long", len(IGNORELIST))

    # Call every 15 minutes
    reactor.callLater(15 * 60, load_ignorelist)


def process_data(data):
    """Callback when we have data to process"""
    try:
        return real_processor(data)
    except Exception as exp:
        common.email_error(exp, data, -1)
    return None


def real_processor(text):
    """Process this product, please"""
    collect = metarcollect.parser(
        text,
        utcnow=common.utcnow(),
        nwsli_provider=NWSLI_PROVIDER,
        ugc_provider={},
    )
    if collect.warnings:
        common.email_error("\n".join(collect.warnings), collect.unixtext)
    jmsgs = collect.get_jabbers(
        SETTINGS.get("pywwa_metar_url", "pywwa_metar_url")
    )
    for jmsg in jmsgs:
        if jmsg[0] in JABBER_MESSAGES:
            continue
        # Hacky here, but get the METAR.XXXX channel to find which site
        # this is.
        skip = False
        channels = jmsg[2].get("channels", [])
        for channel in channels.split(","):
            if (
                channel.startswith("METAR.")
                and channel.split(".")[1] in IGNORELIST
            ):
                LOG.info("IGNORELIST Jabber relay of %s", jmsg[0])
                skip = True
        JABBER_MESSAGES.append(jmsg[0])
        if not skip:
            common.send_message(*jmsg)
    if not common.dbwrite_enabled():
        return collect
    for mtr in collect.metars:
        key = metarcollect.normid(mtr.station_id)
        entry = NWSLI_PROVIDER.get(key)
        if entry is None:
            LOG.info("station: '%s' is unknown to metadata table", key)
            deffer = ASOSDB.runOperation(
                "INSERT into unknown(id, valid) values (%s, %s)",
                (mtr.station_id, mtr.time.replace(tzinfo=timezone.utc)),
            )
            deffer.addErrback(common.email_error, text)
            continue
        deffer = IEMDB.runInteraction(do_db, mtr, entry)
        deffer.addErrback(common.email_error, collect.unixtext)
    return collect


def do_db(txn, mtr, metadata: dict):
    """Do database transaction"""
    # We always want data to at least go to current_log incase we are getting
    # data out of order :/
    iem, res = metarcollect.to_iemaccess(
        txn,
        mtr,
        iemid=metadata["iemid"],
        tzname=metadata["tzname"],
        force_current_log=True,
    )
    if not res:
        LOG.info(
            "INFO: IEMAccess update of %s returned false: %s",
            metadata["id"],
            mtr.code,
        )
        df = ASOSDB.runOperation(
            "INSERT into unknown(id, valid) values (%s, %s)",
            (mtr.station_id, iem.data["valid"]),
        )
        df.addErrback(common.email_error, mtr.station_id)


def cleandb():
    """Reset the JABBER_MESSAGES."""
    LOG.info("cleandb() called...")
    JABBER_MESSAGES.clear()
    # Call Again in 1440 minutes
    reactor.callLater(1440, cleandb)


def ready(_):
    """callback once our database load is done"""
    bridge(process_data)
    cleandb()
    load_ignorelist()
    lc = LoopingCall(
        MESOSITEDB.runInteraction, load_metar_stations, NWSLI_PROVIDER
    )
    lc.start(720, now=False)


@click.command(help=__doc__)
@common.init
def main(*args, **kwargs):
    """Run once at startup"""
    df = MESOSITEDB.runInteraction(load_metar_stations, NWSLI_PROVIDER)
    df.addCallback(ready)
    df.addErrback(LOG.error)



"""Test pywwa.xmpp"""

from unittest import mock

import pytest
from twisted.words.protocols.jabber import jid
from twisted.words.xish import domish, xmlstream

from pywwa import CTX, xmpp


@pytest.fixture(autouse=True)
def reactor():
    """Fixture"""
    return mock.Mock()


def test_raw_data_in():
    """Test the raw data in method."""
    xmpp.raw_data_in(b"hello")


def test_debug():
    """Test the debug method."""
    elem = domish.Element(("jabber:client", "message"))
    xmpp.debug(elem)
    xmpp.debug("hello")


def test_raw_data_out():
    """Test the raw data out method."""
    xmpp.raw_data_out(b"hello")
    xmpp.raw_data_out(" ")


def test_client_authd(reactor):
    """Test the authd method."""
    client = xmpp.JabberClient(reactor, jid.JID("root@localhost"), "secret")
    xs = xmlstream.XmlStream()
    xs.transport = mock.Mock()
    xs.transport.write = mock.Mock()
    client.connected(xs)
    client.authd(xs)


def test_message_processor():
    """Test the message processor."""
    elem = domish.Element(("jabber:client", "message"))
    elem["from"] = "root@localhost"
    elem["to"] = "iembot@localhost"
    elem.addElement("body", content="shutdown")
    xmpp.message_processor(elem)


def test_disabled_xmpp():
    """Test that it works when XMPP is disabled."""
    CTX["disable_xmpp"] = True
    assert isinstance(xmpp.make_jabber_client(), xmpp.NOOPXMPP)
    CTX["disable_xmpp"] = False


def test_illegal_xml():
    """Test that we can remove illegal XML characters."""
    assert xmpp.ILLEGAL_XML_CHARS_RE.sub("", "\003hello") == "hello"


def test_send_message(reactor):
    """Test the sending of messages."""
    client = xmpp.JabberClient(reactor, jid.JID("root@localhost"), "secret")
    client.authenticated = True
    client.xmlstream = xmlstream.XmlStream()
    client.send_message(
        "hello",
        "hello",
        {
            "channels": ["XX", "YY"],
            "t": "x",
            "twitter_media": "http://thiswillfail.lazz/bogus",
        },
    )
    client.disconnect(None)


def test_client(reactor):
    """Test that we can use the JabberClient."""
    client = xmpp.JabberClient(reactor, jid.JID("root@localhost"), "secret")
    assert client.routerjid


def test_make_jabber_client():
    """Test that we can build jabber clients."""
    client = xmpp.make_jabber_client()
    assert client.myjid.resource is not None


def test_iq_ping(reactor):
    """Test the response to an IQ ping request."""
    client = xmpp.JabberClient(reactor, jid.JID("root@localhost"), "secret")
    client.xmlstream = xmlstream.XmlStream()
    client.xmlstream.transport = mock.Mock()
    client.xmlstream.transport.write = mock.Mock()
    ping = domish.Element(("jabber:client", "iq"))
    ping["from"] = "iembot@localhost"
    ping["to"] = "root@localhost"
    ping["type"] = "get"
    ping["id"] = "123"
    ping.addElement("ping", "urn:xmpp:ping")
    client.iq_processor(ping)


def test_iq_version(reactor):
    """Test the response to IQ version request."""
    client = xmpp.JabberClient(reactor, jid.JID("root@localhost/me"), "secret")
    xs = xmlstream.XmlStream()
    xs.transport = mock.Mock()
    xs.transport.write = mock.Mock()
    client.connected(xs)
    client.authd(xs)
    iq = domish.Element(("jabber:client", "iq"))
    iq["from"] = "iembot@localhost"
    iq["to"] = "root@localhost"
    iq["type"] = "get"
    iq["id"] = "123"
    iq.addElement("query", "jabber:iq:version")
    client.iq_processor(iq)



"""Dump some stuff without AFOS PILs"""

# Stdlib
import re

import click
from pyiem.nws.product import TextProduct
from twisted.internet import reactor

# Local
from pywwa import common
from pywwa.database import get_database
from pywwa.ldm import bridge
from pywwa.workflows.afos_dump import write2memcache

CWA = re.compile("^FA(AK|HI|US)2([1-6])$")
MIS = re.compile("^FA(AK|HI|US)20$")
GMET = {
    "LWGE86": "GMTIFR",
    "LWHE00": "GMTTRB",
    "LWIE00": "GMTICE",
}
# https://www.nhc.noaa.gov/archive/recon/readme.txt
NHC = {
    "URPA15": "AHOPA1",
    "URPN15": "AHOPN1",
    "URNT15": "AHONT1",
    "URNT10": "REPNT0",
    "URNT11": "REPNT1",
    "URNT12": "REPNT2",
    "URNT13": "REPNT3",
    "UZNT13": "REPNT3",  # checking with NHC
    "URNT14": "REPNTS",
    "URPA10": "REPPA0",
    "URPA11": "REPPA1",
    "URPA12": "REPPA2",
    "URPA13": "REPPA3",
    "UZPA13": "REPPA3",  # checking with NHC
    "URPN10": "REPPN0",
    "URPN11": "REPPN1",
    "URPN12": "REPPN2",
    "URPN13": "REPPN3",
    "UZPN13": "REPPN3",  # checking with NHC
    "URPN14": "REPPNS",
}


def compute_afos(textprod: TextProduct):
    """Our hackery to assign a fake AFOS pil to a product without AFOS"""
    ttaaii: str = textprod.wmo
    if ttaaii[:4] == "NOXX":
        afos = f"ADM{textprod.source[1:]}"
    elif ttaaii == "CDUS27":
        afos = f"DSM{textprod.source[1:]}"
    elif ttaaii.startswith("UB"):
        afos = "PIREP"
    elif ttaaii in GMET:
        afos = GMET[ttaaii]
    elif MIS.match(ttaaii):
        afos = f"MIS{textprod.source[1:]}"
    elif CWA.match(ttaaii):
        afos = f"CWA{textprod.source[1:]}"
    elif ttaaii[:4] == "FOUS":
        afos = f"FRH{ttaaii[4:]}"
    elif ttaaii[:2] == "FO" and ttaaii[2:4] in [
        "CA",
        "UE",
        "UM",
        "CN",
        "GX",
        "UW",
    ]:
        afos = f"FRHT{ttaaii[4:]}"
    elif textprod.source in ["KNHC", "KWBC"] and ttaaii in NHC:
        afos = NHC[ttaaii]
    else:
        raise Exception(f"Unknown TTAAII {ttaaii} conversion")

    textprod.afos = afos


def really_process_data(txn, data):
    """We are called with a hard coded AFOS PIL"""
    tp = TextProduct(
        data, utcnow=common.utcnow(), ugc_provider={}, parse_segments=False
    )
    if tp.afos is None:
        compute_afos(tp)

    sql = (
        "INSERT into products "
        "(pil, data, source, wmo, entered, bbb) values(%s,%s,%s,%s,%s,%s)"
    )

    sqlargs = (
        tp.afos,
        tp.unixtext,
        tp.source,
        tp.wmo,
        tp.valid.strftime("%Y-%m-%d %H:%M+00"),
        tp.bbb,
    )
    if common.dbwrite_enabled():
        txn.execute(sql, sqlargs)

    # CWA is handled by cwa_parser.py
    if tp.afos[:3] not in ["FRH", "CWA", "DSM"]:
        jmsgs = tp.get_jabbers(
            common.SETTINGS.get("pywwa_product_url", "pywwa_product_url")
        )
        for jmsg in jmsgs:
            common.send_message(*jmsg)

    # We are on a thread, so we need to send this back to the main thread
    reactor.callFromThread(write2memcache, tp.get_product_id(), tp.unixtext)
    return tp


@click.command(help=__doc__)
@common.init
def main(*args, **kwargs):
    """Go Main Go."""
    bridge(really_process_data, dbpool=get_database("afos"))



"""Parse FD text products."""

# 3rd Party
import click
from pyiem.nws.products.fd import parser

# Local
from pywwa import common
from pywwa.database import get_database
from pywwa.ldm import bridge


def processor(txn, text):
    """Protect the realprocessor"""
    prod = parser(text, utcnow=common.utcnow(), ugc_provider={})
    if common.dbwrite_enabled():
        prod.sql(txn)
    if prod.warnings:
        common.email_error("\n".join(prod.warnings), text)
    return prod


@click.command(help=__doc__)
@common.init
@common.disable_xmpp
def main(*args, **kwargs):
    """Go Main Go."""
    bridge(processor, dbpool=get_database("asos", cp_max=2))



"""Ingest BUFR (^IS WMO Header) Data.

This is a work in progress yet, notes and todo as follows:

- [ ] Generate a METAR with a IEM_BUFR flag aset
- Issuer of Identifier (20000+ is WMO program), less is local
- 20006 indicates ICAO would be the last ID

"""

# stdlib
from datetime import timedelta

# 3rd Party
import click
from pybufrkit.decoder import Decoder, generate_bufr_message
from pybufrkit.renderer import NestedJsonRenderer
from pyiem.nws.product import TextProduct
from pyiem.observation import Observation
from pyiem.util import convert_value, utc
from twisted.internet.task import LoopingCall

# Local
from pywwa import LOG, common, get_data_filepath
from pywwa.database import get_database
from pywwa.ldm import bridge

MESOSITEDB = get_database("mesosite", cp_max=1)
IEMDB = get_database("iem", cp_max=1)
WMO2ISO3166 = {}
UNKNOWNS = []
WIGOS = {}
NETWORK = "WMO_BUFR_SRF"
DIRECTS = {
    "004001": "year",
    "004002": "month",
    "004003": "day",
    "004004": "hour",
    "004005": "minute",
    "001015": "station_name",
    "007030": "elevation",
    "005001": "lat",
    "006001": "lon",
}
# life choices were made here
DEPTHMAP = {
    0.1: "4",
    0.2: "8",
    0.4: "16",
    0.5: "20",
    0.8: "32",
    1.0: "40",
    1.6: "64",
    3.2: "128",
}
OCTAS_MAP = {
    0: "CLR",
    1: "FEW",
    2: "FEW",
    3: "SCT",
    4: "SCT",
    5: "BKN",
    6: "BKN",
    7: "BKN",
    8: "OVC",
    9: "VV",
    15: "NSC",
}


def bounds_check(val, low, high, dtype=float):
    """Ensure some QC."""
    if val is None or val < low or val > high:
        return None
    return dtype(val)


def load_xref(txn):
    """Build out WIGOS2IEMID."""
    txn.execute(
        "SELECT wigos, iemid, tzname from stations "
        "WHERE wigos is not null and network = %s",
        (NETWORK,),
    )
    for row in txn.fetchall():
        WIGOS[row["wigos"]] = {
            "iemid": row["iemid"],
            "tzname": row["tzname"],
        }
    LOG.info("Loaded %s WIGOS2IEMID entries", len(WIGOS))


def add_station_cb(iemid, sid):
    """Callback."""
    WIGOS[sid] = {"iemid": iemid, "tzname": None}


def add_station(sid, data, mcursor=None):
    """Add a mesosite station entry.

    Args:
      mcursor (cursor): for testing, a mesosite cursor to use.
    """
    if "lon" not in data or "lat" not in data:
        LOG.info("Skipping %s as no location data", sid)
        WIGOS[sid] = {"iemid": -2}
        return
    if data["lon"] == 0 and data["lat"] == 0:
        LOG.info("Skipping %s as lat=lon=0", sid)
        WIGOS[sid] = {"iemid": -2}
        return
    sname = data.get("sname")
    if sname is None:
        LOG.info("Skipping %s as no station name %s", sid, data)
        WIGOS[sid] = {"iemid": -2}
        return
    sname = sname.replace(",", " ")
    elev = data.get("elevation")
    sql = """
        INSERT into stations(id, wigos, name, network, online,
        geom, elevation, metasite, plot_name, country)
        VALUES (%s, %s, %s, %s, 't',
        ST_Point(%s, %s, 4326), %s, 'f', %s, 'UN') returning iemid
        """
    args = (
        sid,
        sid,
        sname,
        NETWORK,
        data["lon"],
        data["lat"],
        elev,
        sname,
    )

    def _insert(txn):
        """Do the insert."""
        txn.execute(sql, args)
        return txn.fetchone()["iemid"]

    if mcursor is None:
        # no coverage for this :(
        df = MESOSITEDB.runInteraction(_insert)
        df.addCallback(add_station_cb, sid)
        df.addErrback(common.email_error, f"{sid} {data}")
    else:
        add_station_cb(_insert(mcursor), sid)


def process_datalist(txn, prod, datalist, mcursor=None):
    """Do what we can do.

    Args:
      mcursor (cursor): for testing, a mesosite cursor to use.
    """
    data = datalist2iemob_data(datalist, prod.source)
    if not data:
        return
    sid = data.get("sid")
    if sid is None:
        return
    valid = data["valid"]
    # Don't allow products from the distant past
    if valid < (common.utcnow() - timedelta(days=720)):
        LOG.warning(
            "%s %s is from the past %s < %s",
            prod.get_product_id(),
            sid,
            valid,
            common.utcnow() + timedelta(days=720),
        )
        return
    # Don't allow products from the future
    if valid > (common.utcnow() + timedelta(hours=2)):
        LOG.warning(
            "%s %s is from the future %s > %s",
            prod.get_product_id(),
            sid,
            valid,
            common.utcnow() + timedelta(hours=2),
        )
        return
    meta = WIGOS.get(sid, {"iemid": None})
    if meta["iemid"] is None:
        # prevent race condition
        WIGOS[sid] = {"iemid": -1}
        meta = WIGOS[sid]
        add_station(sid, data, mcursor)
    if meta["iemid"] < -1:
        # add station failed, so do nothing
        return
    if meta["iemid"] == -1:
        LOG.info("Skipping %s as iemid is currently -1", sid)
        return
    LOG.debug("%s %s %s %s", sid, meta["iemid"], prod.get_product_id(), data)
    # This likely means that IEM station metadata has yet to sync, so there
    # is no iemaccess entry
    if meta["tzname"] is None:
        LOG.info("Skipping %s as tzname is None", sid)
        return
    ob = Observation(valid=valid, iemid=meta["iemid"], tzname=meta["tzname"])
    ob.data.update(data)
    ob.data["raw"] = f"BUFR: {prod.get_product_id()}"
    ob.save(txn)


def render_members(members, msgs):
    """recursive."""
    for member in members:
        if isinstance(member, list):
            render_members(member, msgs)
        elif isinstance(member, dict):
            if "factor" in member:
                LOG.debug("FACTOR: %s", member["factor"])
                msgs.append(member["factor"])
            if "value" in member:
                if member["value"] is not None:
                    LOG.debug(member)
                    msgs.append(member)
            elif "members" in member:
                render_members(member["members"], msgs)
            else:
                LOG.debug("Dead end %s", member)
        else:
            LOG.debug(member)
            msgs.append(member)


def datalist2iemob_data(datalist, source) -> dict:
    """see what we can do with this."""
    data = {}
    displacement = 0
    depth = 0
    skyc = 1
    skyl = 1
    for msg in datalist:
        LOG.debug("%s %s %s", msg["id"], msg["description"], msg["value"])
        if msg["id"].startswith("001"):
            data[msg["id"]] = msg["value"]
            continue
        if msg["id"] == "007061":  # DEPTH
            depth = msg["value"]
            continue
        if msg["id"] in ["004024", "004025"]:  # TIME PERIOD OR DISPLACEMENT
            displacement = msg["value"] * (60 if msg["id"] == "004024" else 1)
            continue
        # 014016 is net radiation, so sub-optimal
        # 014028 is global solar J m-2
        if msg["id"] == "014028" and displacement == -60:  # net rad
            data["srad_1h_j"] = bounds_check(msg["value"], 0, 10_000_000)
            continue
        if msg["id"] == "012130" and depth is not None:  # soil temperature
            _xref = DEPTHMAP.get(depth)
            if _xref is None:
                LOG.info("Unknown tsoil depth %s", depth)
                continue
            data[f"tsoil_{_xref}in_f"] = bounds_check(
                convert_value(msg["value"], "degK", "degF"),
                -100,
                150,
            )
            continue
        if msg["id"] in DIRECTS:
            data[DIRECTS[msg["id"]]] = msg["value"]
            continue
        # 020011 CLOUD AMOUNT
        if msg["id"] == "020011":
            # 0-8 oktas
            # 9 = sky obscured
            # 15 = sky can't be observed
            data[f"skyc{skyc}"] = OCTAS_MAP.get(msg["value"])
            skyc += 1
            continue
        # 020013 HEIGHT OF BASE OF CLOUD
        if msg["id"] == "020013":
            data[f"skyl{skyl}"] = bounds_check(
                convert_value(msg["value"], "m", "foot"), 0, 100_000, dtype=int
            )
            skyl += 1
            continue
        if msg["id"] == "010004":
            data["pres"] = msg["value"] / 100.0
            continue
        if msg["id"] == "010051":
            data["mslp"] = msg["value"] / 100.0

            continue
        if msg["id"] == "012101":
            data["tmpf"] = bounds_check(
                convert_value(msg["value"], "degK", "degF"),
                -100,
                150,
            )
            continue
        if msg["id"] == "012103":
            data["dwpf"] = bounds_check(
                convert_value(msg["value"], "degK", "degF"),
                -100,
                150,
            )
            continue
        if msg["id"] == "013003":
            data["relh"] = bounds_check(
                msg["value"],
                0,
                100,
            )
            continue
        if msg["id"] == "020001":
            data["vsby"] = bounds_check(
                convert_value(msg["value"], "m", "mile"), 0, 100
            )
            continue
        if msg["id"] == "013011" and displacement == -60:
            data["phour"] = bounds_check(
                convert_value(msg["value"], "mm", "inch"), 0, 100
            )
            continue
        if msg["id"] == "011002" and displacement >= -10:
            data["sknt"] = bounds_check(
                convert_value(msg["value"], "meter per second", "knot"),
                0,
                200,
            )
            continue

        if msg["id"] == "011001" and displacement >= -10:
            data["drct"] = bounds_check(msg["value"], 0, 360)
            continue
        if msg["id"] == "011041" and displacement >= -10:
            data["gust"] = bounds_check(
                convert_value(msg["value"], "meter per second", "knot"), 0, 200
            )
            continue
        if msg["id"] == "011043" and displacement >= -10:
            data["gust_drct"] = bounds_check(msg["value"], 0, 360)
            continue
    if "year" not in data or data["year"] == 0:
        return {}
    data["valid"] = utc(
        data["year"],
        data["month"],
        data["day"],
        data.get("hour", 0),  # Unsure if this is too forgiving
        data.get("minute", 0),
    )
    # Attempt to compute a station ID
    if "001125" in data:
        data["sid"] = (
            f"{data['001125']}-"
            f"{data['001126']}-"
            f"{data['001127']}-"
            f"{data['001128'].decode('ascii', 'ignore').strip()}"
        )
    if "001002" in data:
        ccode = WMO2ISO3166.get(source)
        if ccode is None:
            if source not in UNKNOWNS:
                UNKNOWNS.append(source)
                raise ValueError(f"Unknown WMO2ISO3166 {source}")
            return {}
        data["sid"] = f"0-{ccode}-0-{data['001002']}"
    if "001015" in data:
        data["sname"] = data["001015"].decode("ascii", "ignore").strip()
    # Replace null bytes
    for key in ["sname", "sid"]:
        if key in data:
            data[key] = data[key].replace("\x00", "")
    return data


def gen_bmessages(prod, prodbytes) -> list:
    """Generate BUFR Messages."""
    # ATTM we are sending this to the general text parser, so to set
    # various needed attributes.
    bmsgs = []
    try:
        for bufr_message in generate_bufr_message(Decoder(), prodbytes):
            bmsgs.append(bufr_message)
    except Exception as exp:
        LOG.info(
            "%s %s %s",
            exp,
            prod.get_product_id(),
            prod.source,
        )
    return bmsgs


def bmsg2datalists(bmsg) -> list:
    """Convert BUFR Message to a list of messages."""
    jdata = NestedJsonRenderer().render(bmsg)
    res = []
    for section in jdata:
        for parameter in section:
            if isinstance(parameter["value"], list):
                LOG.debug(
                    "--> %s %s",
                    parameter["name"],
                    type(parameter["value"]),
                )
                for entry in parameter["value"]:
                    LOG.debug("----> entry")
                    if isinstance(entry, list):
                        msgs = []
                        render_members(entry, msgs)
                        res.append(msgs)
                    else:
                        # unexpanded_descriptors
                        LOG.debug(entry)
            else:
                LOG.debug("--> %s %s", parameter["name"], parameter["value"])
    return res


def workflow(prodbytes, cursor=None, mcursor=None):
    """This processes the LDM product provided via ldmbridge.

    Args:
        prodbytes (bytes): The LDM product
        cursor (cursor, optional): The database cursor to use.
        mcursor (cursor, optional): The mesosite database cursor to use.
    """
    pos = prodbytes.find(b"BUFR")
    if pos == -1:
        if prodbytes != b"" and prodbytes.find(b"NIL") == -1:
            LOG.warning("No BUFR found in %s", prodbytes[:100])
        return None, None
    # Gleaning some data from the LDM product can be useful downstream
    header = prodbytes[:pos].decode("ascii")
    prod = TextProduct(
        header, ugc_provider={}, utcnow=common.utcnow(), parse_segments=False
    )
    # Trim LDM product to just the BUFR content
    meat = prodbytes[pos:]
    datalists = []
    # Send to pybufrkit, which generates a list of BUFR messages
    for bmsg in gen_bmessages(prod, meat):
        # Split the BUFR message into lists of parsed data to process
        for datalist in bmsg2datalists(bmsg):
            if not datalist:
                LOG.debug("Prod %s has no datalist", prod.get_product_id())
                continue
            datalists.append(datalist)
            if cursor:
                process_datalist(cursor, prod, datalist, mcursor)
                continue
            # Queue this up for processing
            defer = IEMDB.runInteraction(process_datalist, prod, datalist)
            defer.addErrback(common.email_error, prod.get_product_id())
            defer.addErrback(LOG.warning)
    return prod, datalists


def load_wmo2iso3166():
    """Build cross reference."""
    with open(get_data_filepath("wmo2iso3166.tbl")) as fh:
        for line in fh:
            if line.startswith("#"):
                continue
            tokens = line.split()
            WMO2ISO3166[tokens[0]] = tokens[1]
    LOG.info("Loaded %s WMO2ISO3166 entries", len(WMO2ISO3166))


def ready(_):
    """Callback once we are ready."""
    bridge(workflow, isbinary=True)
    lc = LoopingCall(MESOSITEDB.runInteraction, load_xref)
    df = lc.start(10800, now=False)
    df.addErrback(common.email_error)


@click.command(help=__doc__)
@common.init
@common.disable_xmpp
def main(*args, **kwargs):
    """Go Main Go."""
    load_wmo2iso3166()
    df = MESOSITEDB.runInteraction(load_xref)
    df.addCallback(ready)
    df.addErrback(common.email_error)



"""NLDN from Vaisala, TSV format.

[0] The UALF record type (it will always be 0 for these types of records).
[1] Year, 1970 to 2032.
[2] Month, with January as 1 and December as 12.
[3] Day of the month, 1 to 31.
[4] Hour, 0 to 23.
[5] Minute, 0 to 59.
[6] Second, 0 to 60.
[7] Nanosecond, 0 to 999999999.
[8] Latitude of the calculated location in decimal degrees,
    to 4 decimal places, -90.0 to 90.0.
[9] Longitude of the calculated location in decimal degrees,
    to 4 decimal places, -180.0 to 180.0.
[10] Estimated peak current in kiloamps, -9999 to 9999.
[11] Multiplicity for flash data (1 to 99) or 0 for strokes.
[12] Number of sensors participating in the solution, 2 to 99.
[13] Degrees of freedom when optimizing location, 0 to 99.
[14] The error ellipse angle as a clockwise bearing from 0 degrees north,
     0 to 180.0 degrees.
[15] The error ellipse semi-major axis length in kilometers,
     0.00 to 50.00 km.
[16] The error ellipse semi-minor axis length in kilometers, 0 to 50.0 km.
[17] Chi-squared value from location optimization, 0 to 999.99.
[18] Rise time of the waveform in microseconds, 0 to 99.9.
[19] Peak-to-zero time of the waveform in microseconds, 0 to 999.9.
[20] Maximum rate-of-rise of the waveform in kA/usec
     (will be a negative rate if discharge is negative), -999.9 to 999.9.
[21] Cloud indicator, 1 if Cloud-to-cloud discharge, 0 for Cloud-to-ground.
[22] Angle indicator, 1 if sensor angle data used to compute position,
     0 otherwise.
[23] Signal indicator, 1 if sensor signal data used to compute position,
     0 otherwise.
[24] Timing indicator, 1 if sensor timing data used to compute position,
     0 otherwise.
"""

import math

import click
from pyiem.util import utc

from pywwa import LOG, common
from pywwa.database import get_database
from pywwa.ldm import bridge


def process_data(txn, line: str) -> None:
    """The real processor of the raw data, fun!"""
    if line == "":
        return
    tokens = line.split("\t")
    if len(tokens) != 25:
        LOG.info("Invalid NLDN line: `%s`", line)
        return
    # Only care about CG
    if tokens[21] != "0":
        return
    nanoseconds = int(tokens[7])
    valid = utc(*[int(x) for x in tokens[1:7]]).replace(
        microsecond=nanoseconds // 1000
    )
    table = f"nldn{valid:%Y_%m}"
    # Based on what readNLDNSocket.tcl had
    # The binary decoder maybe should have been dividing this by 10?
    eccentricity = 10.0 * math.sqrt(
        1.0 - (float(tokens[16]) ** 2 / float(tokens[15]) ** 2)
    )
    txn.execute(
        f"INSERT into {table} (valid, geom, signal, multiplicity, "
        "axis, eccentricity, ellipse, chisqr) VALUES (%s, "
        "ST_Point(%s, %s, 4326), %s, %s, %s, %s, %s, %s)",
        (
            valid,
            float(tokens[9]),
            float(tokens[8]),
            float(tokens[10]),  # verified
            int(tokens[11]),  # verified
            int(float(tokens[15])),  # verified
            int(eccentricity),  # verified
            int(float(tokens[14])),  # verified
            int(float(tokens[17])),  # verified
        ),
    )

    return


@click.command(help=__doc__)
@common.init
@common.disable_xmpp
def main(*args, **kwargs):
    """Go Main"""
    bridge(
        process_data,
        isbinary=False,
        product_end=b"\r\n",
        dbpool=get_database("nldn"),
    )



"""Unidata LDM pqact bridge."""

# 3rd Party
from twisted.enterprise.adbapi import ConnectionPool
from twisted.internet import reactor, task

# Local
from pywwa import LOG, ldmbridge, shutdown
from pywwa.common import SETTINGS, email_error


class MyProductIngestor(ldmbridge.LDMProductReceiver):
    """I receive products from ldmbridge and process them 1 by 1 :)"""

    def __init__(self, callback, isbinary=False, **kwargs):
        """Constructor."""
        dedup = SETTINGS.get("pywwa_dedup", "false").lower() == "true"
        super().__init__(isbinary=isbinary, dedup=dedup, **kwargs)
        self.local_callback = callback

    def connectionLost(self, reason=None):
        """called when the connection is lost"""
        LOG.info("STDIN Closed: %s, reactor.stop() in 1 second.", reason)
        shutdown()

    def process_data(self, data):
        """Override below."""
        self.local_callback(data)


def bridge(
    callback,
    dbpool: ConnectionPool = None,
    isbinary=False,
    product_end=None,
    cb2=None,
):
    """Build a pqact bridge.

    Params:
      callback (func): The callback function.
      dbpool (adapi): Database pool to use for callback with transaction.
      isbinary (bool): Is LDM processing binary/bytes vs strings.
      product_end (bytes): A custom product delimiter.
    """

    def dbproxy(data):
        """standardized transaction callback."""
        defer = dbpool.runInteraction(callback, data)
        if cb2 is not None:
            defer.addCallback(cb2)
        defer.addErrback(email_error, data)
        defer.addErrback(LOG.error)

    def nodbproxy(data):
        """Just do a deferred."""
        defer = task.deferLater(reactor, 0, callback, data)
        if cb2 is not None:
            defer.addCallback(cb2)
        defer.addErrback(email_error, data)
        defer.addErrback(LOG.error)

    proto = MyProductIngestor(
        nodbproxy if dbpool is None else dbproxy,
        isbinary=isbinary,
        product_end=product_end,
    )

    # If we have a dbpool, we don't want the reactor to stop until it clears
    # out the queue
    if dbpool is not None:

        def _connectionLost(_reason=None):
            """Shutdown the reactor if we are done."""
            waiting = dbpool.threadpool._queue.qsize()
            LOG.info(
                "STDIN closed. dbpool[dbname=%s] waiting: %s",
                dbpool.connkw.get("dbname"),
                waiting,
            )
            if waiting == 0:
                shutdown()
            else:
                reactor.callLater(10, _connectionLost, _reason)

        proto.connectionLost = _connectionLost

    ldmbridge.LDMProductFactory(proto)
    return proto



"""pyWWA local module."""

# stdlib
import json
import logging
import os

from twisted.internet import reactor

# Default logging configuration
LOG = logging.getLogger("pywwa")
LOG.addHandler(logging.NullHandler())

# Shared configuration
SETTINGS = {}
SETTINGS_FN = "pywwa_settings.json"
# Eventually updated by command line parsing
CTX_DEFAULTS = {
    "disable_email": False,
    "disable_dbwrite": False,
    "replace": False,
    "shutdown_delay": 5,
    "stdout_logging": False,
    "utcnow": None,
}
# hold reference to jabber client
CTX = {
    "JABBER": None,  # Jabber client
    "LCLIST": [],  # List of twisted LoopingCall objects
}
CTX.update(CTX_DEFAULTS)


def shutdown():
    """Shutdown method in given number of seconds."""
    delay = CTX["shutdown_delay"]
    LOG.info("Shutting down in %s seconds...", delay)
    reactor.callLater(delay, reactor.stop)


def get_data_filepath(filename):
    """Return full path to a given filename resource."""
    testfn = os.path.join(os.path.dirname(__file__), "data", filename)
    if not os.path.isfile(testfn):
        raise FileNotFoundError(f"could not locate table file {testfn}")
    return testfn


def load_settings():
    """Attempt to locate our configuration file."""
    dirpaths = [
        os.getcwd(),
        os.environ.get("PYWWA_HOME"),
        os.path.join(os.path.expanduser("~"), "etc"),
        os.path.join(os.path.expanduser("~"), "pyWWA"),
    ]
    for dirpath in dirpaths:
        if dirpath is None:
            continue
        testfn = os.path.join(dirpath, SETTINGS_FN)
        if not os.path.isfile(testfn):
            continue
        with open(testfn, encoding="utf-8") as fh:
            SETTINGS.update(json.load(fh))



"""Test generic_parser."""

# 3rd Party
import pytest
from pyiem.util import utc

# Local
import pywwa
from pywwa.testing import get_example_file
from pywwa.workflows import generic


@pytest.mark.parametrize("database", ["postgis"])
def test_aww(cursor):
    """Test AWW parser."""
    data = get_example_file("AWW.txt")
    pywwa.CTX["utcnow"] = utc(2022, 9, 9, 21, 34)
    prod = generic.process_data({}, cursor, data)
    assert not prod.warnings
    prod = generic.process_data({}, cursor, data.replace("092245", "0z0z0z"))
    assert prod.warnings



"""Unzip the filename given on stdin

called from pqact_iemvs.conf
"""

import os
import random
import string
import sys
import zipfile
from io import BytesIO

BASE = "/mesonet/ldmdata"


def ranstr():
    """used in tmpfn creation"""
    return "".join(
        random.SystemRandom().choice(string.ascii_uppercase + string.digits)
        for _ in range(8)
    )


def main(argv):
    """Read the zipfile from stdin and do important things"""
    data = BytesIO(sys.stdin.buffer.read())
    zf = zipfile.ZipFile(data)
    filename = argv[1]
    # Makedir if it does not exist
    dirname = f"{BASE}/{os.path.dirname(filename)}"
    if not os.path.isdir(dirname):
        os.makedirs(dirname)

    work = []
    for info in zf.infolist():
        fn = info.filename
        tmpfn = f"{dirname}/{ranstr()}_{fn}"
        newfn = f"{dirname}/{fn}"
        with open(tmpfn, "wb") as fp:
            fp.write(zf.read(fn))
        work.append([tmpfn, newfn])

    for [tmpfn, fn] in work:
        os.rename(tmpfn, fn)

    # Write the file
    data.seek(0)
    with open(f"{BASE}/{filename}", "wb") as fp:
        fp.write(data.read())


if __name__ == "__main__":
    main(sys.argv)



"""Exercise the rr7 workflow."""

import pytest

from pywwa.testing import get_example_file
from pywwa.workflows import rr7


@pytest.mark.parametrize("database", ["afos"])
def test_real_process(cursor):
    """Can we process a real RR7 product?"""
    data = get_example_file("SHEF/RR7ZOB.txt")
    rr7.real_process(cursor, data)



"""Test shef_parser."""

import datetime
from functools import partial

# 3rd Party
# pylint: disable=no-member-in-module
from unittest import mock
from zoneinfo import ZoneInfo

import pytest

# Local
from psycopg.errors import DeadlockDetected
from pyiem.util import utc
from twisted.python.failure import Failure

from pywwa import CTX, SETTINGS
from pywwa.testing import get_example_file
from pywwa.workflows import shef

# Ensure that we have a database to work with
shef.build_context()


def run_interaction(cursor, sql, args):
    """Make runOperation go."""
    cursor.execute(sql, args)
    defer = mock.Mock()
    defer.addErrback = lambda x, y: None
    return defer


def sync_workflow(prod, cursor):
    """Common workflow."""
    mydata = shef.restructure_data(prod)
    for sid, data in mydata.items():
        shef.process_site(prod, sid, data)
    # Just exercise the API
    shef.process_accessdb_frontend()

    # Arm the entries for processing
    ts = utc() - datetime.timedelta(days=1)
    for iemid, entry in shef.ACCESSDB_QUEUE.items():
        for localts, record in entry.records.items():
            record["last"] = ts
            shef.write_access_records(
                cursor, [(localts, record)], iemid, entry
            )
    # Just exercise the API
    shef.process_accessdb_frontend()


def test_accessdb_exception():
    """Test some GIGO raises an exception."""
    shef.ACCESSDB_QUEUE[-99] = 0
    shef.process_accessdb_frontend()
    shef.ACCESSDB_QUEUE.pop(-99)


def test_missing_value():
    """Test that this missing value flags a database write to null_ col"""
    shef.LOCS["ALBW3"] = {
        "WI_DCP": {
            "valid": shef.U1980,
            "iemid": -99,
            "tzinfo": ZoneInfo("America/Chicago"),
            "epoc": 1,
            "pedts": 1,
        }
    }
    CTX["utcnow"] = utc(2023, 10, 28, 3, 40)
    shef.process_data(get_example_file("SHEF/RRSNMC.txt"))


def test_exclude():
    """Test the exclusion logic."""
    SETTINGS["pywwa_shef_afos_exclude"] = "RRXXXX"
    shef.build_context()
    payload = get_example_file("SHEF/RR1.txt").replace("RR1DSM", "RRXXXX")
    prod = shef.process_data(payload)
    assert not prod.data


def test_midnight():
    """Test that a midnight report gets moved back one minute."""
    shef.LOCS["AISI4"] = {
        "IA_DCP": {
            "valid": shef.U1980,
            "iemid": -99,
            "tzinfo": ZoneInfo("America/Chicago"),
            "epoc": 1,
            "pedts": 1,
        }
    }
    CTX["utcnow"] = utc(2022, 11, 10, 12)
    payload = get_example_file("SHEF/RR1.txt").replace("1150", "0600")
    shef.process_data(payload)
    assert shef.ACCESSDB_QUEUE[-99].records[utc(2022, 11, 10, 5, 59)]


def test_old_dcpdata():
    """Test that old DCP data is not sent to the database."""
    shef.LOCS["AISI4"] = {
        "IA_DCP": {
            "valid": shef.U1980,
            "iemid": -99,
            "tzinfo": ZoneInfo("America/Chicago"),
            "epoc": 1,
            "pedts": 1,
        }
    }
    CTX["utcnow"] = utc(2022, 11, 10, 12)
    shef.process_data(get_example_file("SHEF/RR1.txt"))
    # rewind to 2021
    CTX["utcnow"] = utc(2021, 11, 10, 12)
    shef.process_data(get_example_file("SHEF/RR1.txt"))


def test_get_localtime():
    """Test when we don't know of a station."""
    tzinfo = ZoneInfo("America/Chicago")
    assert shef.get_localtime("XX_DCP123456", tzinfo) == tzinfo


def test_231027_rr8msr_novariable():
    """Test this found in the wild."""
    CTX["utcnow"] = utc(2023, 10, 19, 15)
    prod = shef.process_data(get_example_file("SHEF/RR8MSR.txt"))
    assert not prod.data


def test_empty_product():
    """Exercise the API."""
    CTX["utcnow"] = utc(2023, 9, 15, 13)
    prod = shef.process_data(get_example_file("SHEF/RR7KRF.txt"))
    assert not prod.data and not prod.warnings


def test_enter_unknown_populates():
    """Test that processing a SHEF file with an unknown actually populates."""
    CTX["utcnow"] = utc(2023, 9, 26, 18)
    payload = get_example_file("SHEF/RR8KRF.txt").replace("PLMW4", "ZZZW4")
    shef.process_data(payload)
    assert shef.UNKNOWN["ZZZW4"]


@pytest.mark.parametrize("database", ["hads"])
def test_enter_unknown(cursor):
    """Insert unknown station."""
    shef.enter_unknown(cursor, "AMSI4", "XX", "XX")
    shef.enter_unknown(cursor, "XX_DCP123456", "XX", "XX")


@pytest.mark.parametrize("database", ["hads"])
def test_230926_rr8krf(cursor):
    """Test a smallint error this found in production."""
    CTX["utcnow"] = utc(2023, 9, 26, 18)
    prod = shef.process_data(get_example_file("SHEF/RR8KRF.txt"))
    for element in prod.data:
        args = (
            element.station,
            element.valid,
            element.varname(),
            element.num_value,
            element.depth,
            element.unit_convention,
            element.dv_interval,
            element.qualifier,
        )
        shef.insert_raw_inbound(cursor, args)


def test_bad_element_in_current_queue():
    """Test GIGO on current_queue."""
    shef.CURRENT_QUEUE.clear()
    shef.CURRENT_QUEUE["HI|BYE|HI"] = {"dirty": True}
    shef.save_current()


@pytest.mark.parametrize("database", ["iem"])
def test_omit_report(cursor):
    """Test that the report is omitted..."""
    shef.CURRENT_QUEUE.clear()
    CTX["utcnow"] = utc(2023, 5, 12, 18)
    cursor.execute(
        "INSERT into stations(id, network, tzname, iemid) VALUES "
        "('CORM6', 'MS_DCP', 'America/Chicago', -99)"
    )

    shef.LOCS["CORM6"] = {
        "MS_COOP": {
            "valid": shef.U1980,
            "iemid": -99,
            "tzinfo": ZoneInfo("America/Chicago"),
            "epoc": 1,
            "pedts": 1,
        }
    }
    prod = shef.process_data(get_example_file("CORM6.txt"))
    sync_workflow(prod, cursor)
    cursor.execute(
        "SELECT report, max_tmpf from summary_2023 "
        "where iemid = -99 and day = '2023-05-12'"
    )
    row = cursor.fetchone()
    assert row["max_tmpf"] == 83
    assert row["report"] is not None
    ans = row["report"]
    prod = shef.process_data(get_example_file("CORM6_RTP.txt"))
    sync_workflow(prod, cursor)
    cursor.execute(
        "SELECT report, max_tmpf from summary_2023 "
        "where iemid = -99 and day = '2023-05-12'"
    )
    row = cursor.fetchone()
    assert row["max_tmpf"] == 84
    assert row["report"] == ans

    CTX["ACCESSDB"].runOperation = partial(run_interaction, cursor)
    assert shef.save_current() == 7
    assert shef.save_current() == 0


def test_process_site_eb():
    """Test that the errorback works without any side effects."""
    CTX["utcnow"] = utc(2017, 8, 15, 14)
    shef.process_data(get_example_file("RR7.txt"))
    entry = shef.ACCESSDB_ENTRY(
        station="", network="", tzinfo=ZoneInfo("America/Chicago"), records={}
    )
    record = {"data": {}, "last": utc(), "product_id": ""}
    shef.write_access_records_eb(
        Failure(Exception("Hi Daryl")), [(utc(), record)], 0, entry
    )
    shef.write_access_records_eb(
        Failure(DeadlockDetected()), [(utc, record)], 0, entry
    )


def test_checkvars():
    """Excerise the checkvars logic with the RR1.txt example"""
    # Define an entry with two networks so to make life choices
    shef.LOCS["AISI4"] = {
        "IA_COOP": {
            "valid": shef.U1980,
            "iemid": -99,
            "tzinfo": ZoneInfo("America/Chicago"),
            "epoc": 1,
            "pedts": 1,
        },
        "ISUSM": {
            "valid": shef.U1980,
            "iemid": -99,
            "tzinfo": ZoneInfo("America/Chicago"),
            "epoc": 1,
            "pedts": 1,
        },
    }
    payload = get_example_file("SHEF/RR1.txt")
    CTX["utcnow"] = utc(2022, 11, 10, 12)
    for repl in ["TX", "HG", "SF", "PPH"]:
        shef.process_data(payload.replace("/TX", f"/{repl}"))
    shef.process_data(payload.replace("RR1", "RR3"))
    shef.LOCS["AISI4"].pop("IA_COOP")
    shef.process_data(payload)


def test_restructure_data_eightchar_id():
    """Test that we omit a greather than 8 char station ID."""
    CTX["utcnow"] = utc(2017, 8, 15, 13)
    prod = shef.process_data(get_example_file("RR7.txt"))
    res = shef.restructure_data(prod)
    assert all(len(x) <= 8 for x in res)


def test_restructure_data_future():
    """Ensure that data from the future stays out!"""
    CTX["utcnow"] = utc(2017, 8, 14)
    prod = shef.process_data(get_example_file("RR7.txt"))
    res = shef.restructure_data(prod)
    assert len(res) == 0
    CTX["utcnow"] = utc()
    res = shef.restructure_data(prod)
    assert len(res) == 0


def test_process_data():
    """Test that we can really_process a product!"""
    CTX["utcnow"] = utc(2020, 9, 15)
    prod = shef.process_data(get_example_file("RR7.txt"))
    assert prod.get_product_id() == "202009151244-KWOH-SRUS27-RRSKRF"


def test_mydict():
    """Test what the custom dict class of MyDict can do."""
    d = shef.MyDict()
    d["HI"] = "HI"
    assert d["HI"] == "HI"
    assert d["TAIRRXZ"] == "max_tmpf"
    assert d["QQQQQQQ"] == ""


@pytest.mark.parametrize("database", ["mesosite"])
def test_load_stations(cursor):
    """Test our load_stations function."""
    shef.UNKNOWN["_X_X_"] = True
    # Create an entry with a bad timezone
    cursor.execute(
        "INSERT into stations(id, network, tzname) VALUES "
        "('_X_X_', 'XX_DCP', 'XXX')"
    )
    cursor.execute(
        "INSERT into stations(id, network, tzname) VALUES "
        "('_Y_Y_', 'XX_DCP', '')"
    )
    shef.load_stations(cursor)
    assert isinstance(shef.LOCS, dict)
    # Remove station from database
    cursor.execute(
        "DELETE from stations where id = '_X_X_' and network = 'XX_DCP'"
    )
    shef.load_stations(cursor)
    assert "_X_X_" not in shef.LOCS


def test_load_stations_fe():
    """Test the blocking frontend."""
    shef.load_stations_fe(True)
    shef.load_stations_fe(False)


def test_api():
    """Exercise the API."""
    shef.log_database_queue_size()


def test_lc_proxy_failure():
    """Call something that fails."""

    def _failure():
        raise Exception("Hi Daryl")

    shef.lc_proxy(_failure)



"""Test pirep_parser."""

# 3rd Party
import pytest

from pywwa.testing import get_example_file

# Local
from pywwa.workflows import pirep


@pytest.mark.parametrize("database", ["postgis"])
def test_real_parser(cursor):
    """Test that we can parse PIREPs!"""
    pirep.real_parser(cursor, get_example_file("PIREP.txt"))
    pirep.cleandb()


@pytest.mark.parametrize("database", ["mesosite"])
def test_load_locs(cursor):
    """Test that we can load our stations table."""
    pirep.load_locs(cursor)



"""ASOS Daily Summary Message Parser ingestor"""

from zoneinfo import ZoneInfo

# 3rd Party
import click
from pyiem.nws.products.dsm import parser

# Local
from pywwa import LOG, common
from pywwa.database import get_database, get_dbconnc
from pywwa.ldm import bridge

# database timezones to cache
TIMEZONES = {}
STATIONS = {}


def load_stations(txn):
    """load station metadata to build a xref."""
    txn.execute("SELECT id, tzname from stations where network ~* 'ASOS'")
    for row in txn:
        # we need four char station IDs
        station = row["id"] if len(row["id"]) == 4 else f"K{row['id']}"
        tzname = row["tzname"]
        if tzname not in TIMEZONES:
            try:
                TIMEZONES[tzname] = ZoneInfo(tzname)
            except Exception as exp:
                LOG.info("ZoneInfo does not like tzname: %s %s", tzname, exp)
                TIMEZONES[tzname] = ZoneInfo("UTC")
        STATIONS[station] = TIMEZONES[tzname]


def real_parser(txn, data):
    """Please process some data"""
    prod = parser(data, utcnow=common.utcnow(), ugc_provider={})
    prod.tzlocalize(STATIONS)
    if common.dbwrite_enabled():
        prod.sql(txn)
    if prod.warnings:
        common.email_error("\n".join(prod.warnings), data)


@click.command(help=__doc__)
@common.init
def main(*args, **kwargs):
    """build things up."""
    # sync
    pgconn, cursor = get_dbconnc("mesosite")
    load_stations(cursor)
    pgconn.close()
    bridge(real_parser, dbpool=get_database("iem"))



"""SHEF product ingestor."""

# stdlib
import datetime
import random
import re
from collections import namedtuple
from zoneinfo import ZoneInfo

# 3rd Party
# pylint: disable=no-name-in-module
import click
from psycopg.errors import DeadlockDetected
from pyiem.models.shef import SHEFElement
from pyiem.nws.products.shef import parser
from pyiem.observation import Observation
from pyiem.util import convert_value, utc
from twisted.internet import reactor
from twisted.internet.task import LoopingCall, deferLater

# Local
from pywwa import CTX, LOG, SETTINGS, common
from pywwa.database import get_database, get_dbconnc
from pywwa.ldm import bridge

# A list of AFOS IDs that we will exclude
AFOS_EXCLUDE = []

# a form for IDs we will log as unknown
NWSLIRE = re.compile("^[A-Z]{4}[0-9]$")

# AFOS IDs for which we do not save the raw report for
SKIP4REPORT = re.compile("^(HYD|RTP)")

# stations we don't know about
UNKNOWN = {}
# station metadata
LOCS = {}
# a queue for saving database IO
CURRENT_QUEUE = {}
# Data structure to hold potential writes to IEMAccess
# [iemid] -> ACCESSDB_ENTRY -> records -> localts -> {}
ACCESSDB_QUEUE = {}
ACCESSDB_ENTRY = namedtuple(
    "ACCESSDB_ENTRY",
    [
        "station",
        "network",
        "tzinfo",
        "records",
    ],
)
U1980 = utc(1980)
# Networks that can come via SHEF backdoors
DOUBLEBACKED_NETWORKS = ["ISUSM", "IA_RWIS"]
P1H = datetime.timedelta(hours=1)
P60D = datetime.timedelta(days=60)


DIRECTMAP = {
    "EPDZ": "et_inch",
    "HGIZ": "rstage",
    "HPIZ": "rstage",
    "HTIZ": "rstage",
    "PAIZ": "pres",
    "PPHZ": "phour",
    "PPDZ": "pday",
    "PPPZ": "pday",
    "PCIZ": "pcounter",
    "QRIZ": "discharge",
    "QTIZ": "discharge",
    "RWIZ": "srad",
    "SDIZ": "snowd",
    "SFDZ": "snow",
    "SWIZ": "snoww",
    "TDIZ": "dwpf",
    "TAIZ": "tmpf",
    "TAIN": "min_tmpf",
    "TAIX": "max_tmpf",
    "TWIZ": "water_tmpf",
    "UDIZ": "drct",
    "UGIZ": "gust",
    "UPIZ": "gust",
    "UPHZ": "gust",
    "URHZ": "max_drct",
    "URIZ": "max_drct",
    "USIZ": "sknt",
    "VBIZ": "battery",
    "XRIZ": "relh",
    "XVIZ": "vsby",
}


class MyDict(dict):
    """
    Customized dictionary class
    """

    def __getitem__(self, key):
        """
        Over-ride getitem so that the sql generated is proper
        """
        val = self.get(key)
        if val is not None:
            return val
        # Logic to convert this key into something the iem can handle
        # Our key is always at least 6 chars!
        pei = f"{key[:3]}{key[5]}"
        if pei in DIRECTMAP:
            self.__setitem__(key, DIRECTMAP[pei])
            return DIRECTMAP[pei]
        LOG.info("Can not map var %s", key)
        self.__setitem__(key, "")
        return ""


MAPPING = MyDict()


def load_stations_fe(blocking=False):
    """Load station metadata, sync is if we use defers."""
    if blocking:
        conn, cursor = get_dbconnc("mesosite")
        load_stations(cursor)
        cursor.close()
        conn.close()
    else:
        dbpool = get_database("mesosite", cp_max=1)
        df = dbpool.runInteraction(load_stations)
        df.addErrback(common.email_error)
        df.addBoth(lambda _: dbpool.close())


def load_stations(txn):
    """Load station metadata

    We need this information as we can't reliably convert a station ID found
    in a SHEF encoded product to a network that is necessary to update IEM
    Access.

    Args:
      txn: a database transaction
    """
    LOG.info("load_stations called...")
    # When tzname is null, we likely have an incomplete entry
    txn.execute(
        """
        SELECT id, s.iemid, network, tzname, a.value as pedts from stations s
        LEFT JOIN station_attributes a on (s.iemid = a.iemid and
        a.attr = 'PEDTS') WHERE (network ~* 'COOP'
        or network ~* 'DCP' or network = ANY(%s))
        and tzname is not null ORDER by network ASC
        """,
        (DOUBLEBACKED_NETWORKS,),
    )

    # A sentinel to know if we later need to remove things in the case of a
    # station that got dropped from a network
    epoc = utc().microsecond
    for row in txn.fetchall():
        stid = row["id"]
        iemid = row["iemid"]
        network = row["network"]
        try:
            tzinfo = ZoneInfo(row["tzname"])
        except Exception:
            LOG.info("ZoneInfo does not like tzname: %s", row["tzname"])
            tzinfo = ZoneInfo("UTC")
        pedts = row["pedts"]
        if stid in UNKNOWN:
            LOG.info("  station: %s is no longer unknown!", stid)
            UNKNOWN.pop(stid)
        metadata = LOCS.setdefault(stid, {})
        if network not in metadata:
            metadata[network] = {
                "valid": U1980,
                "iemid": iemid,
                "tzinfo": tzinfo,
                "epoc": epoc,
                "pedts": pedts,
            }
        else:
            metadata[network]["epoc"] = epoc

    # Now we find things that are outdated, note that other code can add
    # placeholders that can get zapped here.
    for stid in list(LOCS):
        for network in list(LOCS[stid]):
            if LOCS[stid][network]["epoc"] != epoc:
                LOG.info(" sid: %s no longer in network: %s", stid, network)
                LOCS[stid].pop(network)
        if not LOCS[stid]:
            LOG.info(" sid: %s removed completely", stid)
            LOCS.pop(stid)

    LOG.info("loaded %s stations", len(LOCS))


def restructure_data(prod):
    """Create a nicer data structure for future processing."""
    mydata = {}
    # Step 1: Restructure and do some cleaning
    old = []
    utcnow = common.utcnow()
    for se in prod.data:
        # We don't want any non-report / forecast data, missing data
        if se.type != "R":
            continue
        # We don't care about data in the future!
        if se.valid > (utcnow + P1H):
            continue
        if se.valid < (utcnow - P60D):
            if se.station in old:
                continue
            LOG.info("Rejecting old data %s %s", se.station, se.valid)
            old.append(se.station)
            continue
        s_data = mydata.setdefault(se.station, {})
        st_data = s_data.setdefault(se.valid, [])

        # TODO handling of DV properly
        st_data.append(se)
    return mydata


def enter_unknown(cursor, sid, product_id, network):
    """
    Enter some info about a site ID we know nothing of...
    @param sid string site id
    @param product_id string
    @param network string of the guessed network
    """
    # Eh, lets not care about non-5 char IDs
    if NWSLIRE.match(sid) is None:
        return
    cursor.execute(
        "INSERT into unknown(nwsli, product, network) values (%s, %s, %s)",
        (sid, product_id, network),
    )


def checkvars(myvars):
    """
    Check variables to see if we have a COOP or DCP site
    """
    for v in myvars:
        # Definitely DCP
        if v[:2] in ["HG"]:
            return False
        if v[:2] in ["SF", "SD"]:
            return True
        if v[:3] in ["PPH"]:
            return False
    return False


def save_current() -> int:
    """update the database current_shef table

    It turns out that our database can't handle this fast enough in realtime,
    so we aggregate some
    """
    cnt = 0
    skipped = 0
    for k, mydict in CURRENT_QUEUE.items():
        if not mydict["dirty"]:
            skipped += 1
            continue
        cnt += 1
        # GH215 somehow sid has a | in it, maybe
        (sid, varname, _depth) = k.split("|", 2)
        # This is unlikely, but still GIGO sometimes :/
        if len(varname) != 7:
            LOG.info("Got varname of '%s' somehow? %s", varname, mydict)
            continue

        d2 = CTX["ACCESSDB"].runOperation(
            "INSERT into current_shef(station, valid, physical_code, "
            "duration, source, type, extremum, probability, value, depth, "
            "dv_interval, unit_convention, qualifier, product_id) "
            "values (%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s)",
            (
                sid,
                mydict["valid"],
                varname[:2],
                varname[2],
                varname[3],
                varname[4],
                varname[5],
                varname[6],
                mydict["value"],
                mydict["depth"],
                mydict["dv_interval"],
                mydict["unit_convention"],
                mydict["qualifier"],
                mydict["product_id"],
            ),
        )
        d2.addErrback(common.email_error, "")
        mydict["dirty"] = False

    LOG.info("processed %s entries, %s skipped", cnt, skipped)
    return cnt


def get_localtime(sid, ts):
    """Compute the local timestamp for this location"""
    if sid not in LOCS:
        return ts
    _network = list(LOCS[sid])[0]
    return ts.astimezone(LOCS[sid][_network]["tzinfo"])


def get_network(prod, sid, data: list[SHEFElement]) -> str:
    """Logic for figuring out network in face of ambiguity.

    Note: This sid should already be in LOCS, so we are only either taking
    the single entry or picking between DCP and COOP variants.
    """
    networks = list(LOCS[sid].keys())
    # This is the best we can hope for
    if len(networks) == 1:
        return networks[0]
    # Our simple determination if the site is a COOP site
    is_coop = False
    varnames = [se.varname() for se in data]
    if prod.afos[:3] == "RR3":
        is_coop = True
    elif prod.afos[:3] in ["RR1", "RR2"] and checkvars(varnames):
        LOG.info(
            "Guessing COOP? %s %s %s", sid, prod.get_product_id(), varnames
        )
        is_coop = True
    pnetwork = "COOP" if is_coop else "DCP"
    # filter networks now
    for network in networks:
        if network.find(pnetwork) > 0:
            return network
    # Throw our hands up in the air
    return networks[0]


def process_site(prod, sid, data):
    """Consumption of rectified data."""
    # Order the timestamps so that we process the newest data last, so that
    # all obs are potentially processed through iemaccess
    times = list(data.keys())
    times.sort()
    for tstamp in times:
        process_site_time(prod, sid, tstamp, data[tstamp])


def insert_raw_inbound(cursor, args) -> int:
    """Do the database insertion."""
    cursor.execute(
        "INSERT into raw_inbound (station, valid, key, value, depth, "
        "unit_convention, dv_interval, qualifier) "
        "VALUES(%s, %s, %s, %s, %s, %s, %s, %s)",
        args,
    )
    return cursor.rowcount


def update_current_queue(element: SHEFElement, product_id: str):
    """Update CURRENT_QUEUE with new data."""
    # We only want observations
    if element.type != "R":
        return
    args = (
        element.station,
        element.valid,
        element.varname(),
        element.num_value,
        element.depth,
        element.unit_convention,
        element.dv_interval,
        element.qualifier,
    )
    defer = CTX["HADSDB"].runInteraction(insert_raw_inbound, args)
    defer.addErrback(common.email_error, f"prod: {product_id}")
    defer.addErrback(LOG.error)

    key = f"{element.station}|{element.varname()}|{element.depth}"
    cur = CURRENT_QUEUE.setdefault(
        key, dict(valid=element.valid, value=element.num_value, dirty=True)
    )
    if element.valid < cur["valid"]:
        return
    cur["valid"] = element.valid
    cur["depth"] = element.depth
    cur["value"] = element.num_value
    cur["dv_interval"] = element.dv_interval
    cur["qualifier"] = element.qualifier
    cur["unit_convention"] = element.unit_convention
    cur["product_id"] = product_id
    cur["dirty"] = True


def process_site_time(prod, sid, ts, elements: list[SHEFElement]):
    """Ingest for IEMAccess."""
    network = get_network(prod, sid, elements)
    if network is None or network in DOUBLEBACKED_NETWORKS:
        return
    localts = get_localtime(sid, ts)
    # This should always work!
    metadata = LOCS[sid][network]
    # Do not send DCP sites with old data to IEMAccess
    if network.find("_DCP") > 0 and localts < metadata["valid"]:
        return
    metadata["valid"] = localts

    # Okay, time for a hack, if our observation is at midnight!
    if localts.hour == 0 and localts.minute == 0:
        localts -= datetime.timedelta(minutes=1)

    record = ACCESSDB_QUEUE.setdefault(
        metadata["iemid"],
        ACCESSDB_ENTRY(
            station=sid,
            network=network,
            tzinfo=metadata["tzinfo"],
            records={},
        ),
    ).records.setdefault(
        localts,
        {
            "last": U1980,
            "data": {},
            "product_id": prod.get_product_id(),
        },
    )
    # Update last
    record["last"] = utc()

    report = None
    afos = prod.afos
    for se in elements:
        if se.narrative:
            report = se.narrative
        if (
            report is None
            and afos is not None
            and SKIP4REPORT.match(afos[:3]) is None
        ):
            report = se.raw
        varname = se.varname()
        iemvar = MAPPING[varname]
        if iemvar == "":  # or (iemvar == "rstage" and pedts is not None):
            continue
        # We generally are in english units, this also converts the wind
        # direction to the full degrees
        val = se.to_english()
        if val is None:
            # Behold, glorious hack here to force nulls into the summary
            # database that uses coerce
            record["data"][f"null_{iemvar}"] = None
        record["data"][iemvar] = val
        if iemvar in ["pday", "snow", "snowd"]:
            # Prevent negative numbers
            if val is not None and val < 0:
                record["data"][iemvar] = 0
        if iemvar in ["sknt", "gust"] and val is not None:
            # mph to knots :/
            val = convert_value(val, "mile / hour", "knot")
        if network.find("_COOP") > 0:
            # Save COOP 'at-ob' temperature into summary table
            if iemvar == "tmpf":
                record["data"]["coop_tmpf"] = val
            # Save observation time into the summary table
            if iemvar in [
                "tmpf",
                "max_tmpf",
                "min_tmpf",
                "pday",
                "snow",
                "snowd",
            ]:
                record["data"]["coop_valid"] = localts

    record["data"]["raw"] = prod.get_product_id()
    record["data"]["report"] = report


def write_access_records(
    accesstxn, records: list, iemid, entry: ACCESSDB_ENTRY
):
    """Batch the records to to prevent deadlocks, maybe!"""
    for localts, record in records:
        write_access_record(accesstxn, record, iemid, localts, entry)


def write_access_record(
    accesstxn, record: dict, iemid, localts, entry: ACCESSDB_ENTRY
):
    """The batched database write."""
    iemob = Observation(iemid=iemid, valid=localts, tzname=entry.tzinfo.key)
    iemob.data.update(record["data"])
    iscoop = entry.network.find("_COOP") > 0
    iemob.save(accesstxn, force_current_log=iscoop)


def log_database_queue_size():
    """Log how much backlog has accumulated."""
    LOG.info(
        "dbpool queuesz[hads:%s, access:%s]",
        CTX["HADSDB"].threadpool._queue.qsize(),  # skipcq: PYL-W0212
        CTX["ACCESSDB"].threadpool._queue.qsize(),  # skipcq: PYL-W0212
    )


def write_access_records_eb(err, records: list, iemid, entry: ACCESSDB_ENTRY):
    """Errorback from process_site transaction."""
    if isinstance(err.value, DeadlockDetected):
        jitter = random.randint(0, 30)
        LOG.info(
            "Database Deadlock: %s[%s], retrying in %s seconds",
            entry.station,
            entry.network,
            jitter,
        )
        df = deferLater(
            reactor,
            jitter,
            CTX["ACCESSDB"].runInteraction,
            write_access_records,
            records,
            iemid,
            entry,
        )
        df.addErrback(common.email_error)
        return
    common.email_error(err, f"write_access_entry({entry.station}) got {err}")


def process_data(text):
    """Callback when text is received."""
    prod = parser(text, utcnow=common.utcnow(), ugc_provider={})
    if prod.warnings:
        common.email_error("\n".join(prod.warnings), prod.unixtext)
    if prod.afos in AFOS_EXCLUDE:
        LOG.info("Skipping AFOS: %s due to in AFOS_EXCLUDE", prod.afos)
        prod.data = []
        return prod
    if not prod.data:
        return prod
    product_id = prod.get_product_id()
    # Update CURRENT_QUEUE
    time_threshold = common.utcnow() + P1H
    for element in [e for e in prod.data if e.valid < time_threshold]:
        update_current_queue(element, product_id)
    # Create a nicer data structure
    mydata = restructure_data(prod)
    # Chunk thru each of the sites found and do work.
    for sid, data in mydata.items():
        # Can't send unknown sites to iemaccess
        if sid in UNKNOWN:
            continue
        if sid in LOCS:
            process_site(prod, sid, data)
        else:
            UNKNOWN[sid] = True
            if NWSLIRE.match(sid) is not None:
                LOG.info("Unknown NWSLIish site: %s %s", sid, product_id)
                CTX["HADSDB"].runInteraction(
                    enter_unknown, sid, product_id, ""
                )
    return prod


def process_accessdb_frontend():
    """Catch exceptions so that Looping call keeps going."""
    try:
        process_accessdb()
    except Exception as exp:
        LOG.exception(exp)


def process_accessdb():
    """Queue up work to do."""
    threshold = utc() - datetime.timedelta(minutes=5)
    for iemid, entry in ACCESSDB_QUEUE.items():
        records = []
        for localts in list(entry.records.keys()):
            # Allow a window of time to accumulate data prior to writing
            if entry.records[localts]["last"] > threshold:
                continue
            # Get a reference and delete it from the dict
            records.append((localts, entry.records.pop(localts)))
        if records:
            df = CTX["ACCESSDB"].runInteraction(
                write_access_records,
                records,
                iemid,
                entry,
            )
            df.addErrback(
                write_access_records_eb,
                records,
                iemid,
                entry,
            )
            df.addErrback(common.email_error)


def build_context():
    """Build up things necessary for this to run."""
    if "pywwa_shef_afos_exclude" in SETTINGS:
        AFOS_EXCLUDE.extend(SETTINGS["pywwa_shef_afos_exclude"].split(","))
        LOG.info("AFOS_EXCLUDE: %s", AFOS_EXCLUDE)
    load_stations_fe(True)

    # Construct the needed database pools
    CTX["ACCESSDB"] = get_database("iem", module_name="psycopg", cp_max=20)
    CTX["HADSDB"] = get_database("hads", module_name="psycopg", cp_max=20)

    # Add some looping calls
    CTX["LCLIST"].extend(
        [
            # write out cached obs to IEMAccess
            LoopingCall(lc_proxy, save_current).start(131, now=False),
            # Run a logged diagnostic on how busy the database pools are
            LoopingCall(lc_proxy, log_database_queue_size).start(
                61, now=False
            ),
            # Reload the station table every 12 hours
            LoopingCall(lc_proxy, load_stations_fe).start(
                60 * 60 * 12, now=False
            ),
            # Process entries in ACCESSDB_QUEUE
            LoopingCall(lc_proxy, process_accessdb_frontend).start(
                15, now=False
            ),
        ]
    )


def lc_proxy(f):
    """Ensure that we don't have an uncaught exception."""
    try:
        f()
    except Exception as exp:
        common.email_error(exp, f.__name__)
        LOG.exception(exp)


@click.command(help=__doc__)
@click.option("--custom-arg", "-c", type=str, help="Differentiate pqact job")
@common.init
@common.disable_xmpp
def main(*args, **kwargs):
    """We startup."""
    build_context()
    bridge(process_data)



"""Get station updates from IEM Webservice."""

# Third Party
import httpx

from pywwa.database import get_dbconnc


def main():
    """Go Main Go."""
    pgconn, cursor = get_dbconnc("mesosite")

    req = httpx.get("http://mesonet.agron.iastate.edu/json/stations.php")
    jdata = req.json()
    for site in jdata["stations"]:
        if site["network"].find("ASOS") == -1:
            continue
        # Hmmm
        if site["wfo"] is None:
            site["wfo"] = ""
        cursor.execute(
            "SELECT iemid from stations WHERE id = %s and network = %s",
            (site["id"], site["network"]),
        )
        if cursor.rowcount == 0:
            cursor.execute(
                "INSERT into stations(id, network) values (%s, %s)",
                (site["id"], site["network"]),
            )
        giswkt = f"SRID=4326;POINT({site['lon']} {site['lat']})"
        cursor.execute(
            """
            UPDATE stations SET name = %s, state = %s, elevation = %s,
            geom = %s, county = %s, wfo = %s
            WHERE id = %s and network = %s
            """,
            (
                site["name"],
                site["state"],
                site["elevation"],
                giswkt,
                site["county"],
                site["wfo"],
                site["id"],
                site["network"],
            ),
        )

    cursor.close()
    pgconn.commit()


if __name__ == "__main__":
    main()



"""SPENES product ingestor"""

# stdlib
import re

import click
from pyiem.nws import product

# Local
from pywwa import common
from pywwa.database import get_database
from pywwa.ldm import bridge


def real_process(txn, raw):
    """Do work please"""
    _sqlraw = raw.replace("\015\015\012", "\n")
    prod = product.TextProduct(raw, ugc_provider={}, utcnow=common.utcnow())

    product_id = prod.get_product_id()
    if common.dbwrite_enabled():
        txn.execute(
            "INSERT into text_products(product_id) values (%s)",
            (product_id,),
        )

    tokens = re.findall("ATTN (WFOS|RFCS)(.*)", raw)
    channels = []
    for tpair in tokens:
        for center in re.findall(r"([A-Z]+)\.\.\.", tpair[1]):
            channels.append(f"SPENES.{center}")
    baseurl = common.SETTINGS.get("pywwa_product_url", "pywwa_product_url")
    xtra = {"product_id": product_id}
    xtra["channels"] = ",".join(channels)
    body = (
        "NESDIS issues Satellite Precipitation Estimates "
        f"{baseurl}?pid={product_id}"
    )
    xtra["twitter"] = body
    htmlbody = (
        f"<p>NESDIS issues <a href='{baseurl}?pid={product_id}'>"
        "Satellite Precipitation Estimates</a></p>"
    )
    common.send_message(body, htmlbody, xtra)


@click.command(help=__doc__)
@common.init
def main(*args, **kwargs):
    """Go Main Go."""
    bridge(real_process, dbpool=get_database("postgis"))



"""Some magic to feed the SEL, SAW, and WWP products into the LDM bridge."""

import sys
import time


def main(argv):
    """Go Main Go."""
    with open(argv[1], "rb") as fh:
        payload = fh.read()
    while True:
        sys.stdout.buffer.write(payload)
        time.sleep(1)


if __name__ == "__main__":
    main(sys.argv)



"""Review our NWWS-OI file for deltas."""

# stdlib
import sys

# third party
import httpx
from pyiem.nws.product import TextProduct


def main(argv):
    """Go Main Go."""
    fn = argv[1]
    with open(fn, "rb") as fh:
        for token in fh.read().decode("ascii", "ignore").split("\003"):
            if len(token) < 10:
                continue
            prod = TextProduct(token, ugc_provider={})
            if prod.afos is None:
                continue
            pid = prod.get_product_id()
            uri = f"https://mesonet.agron.iastate.edu/api/1/nwstext/{pid}"
            req = httpx.get(uri)
            if req.status_code != 200:
                print(uri)


if __name__ == "__main__":
    main(sys.argv)



"""Hackery."""

import matplotlib.colors as mpcolors

# Make colormap (based off of
# http://cimss.ssec.wisc.edu/goes/visit/water_vapor_enhancement.html):
colors = [
    "#000000",
    "#00FFFF",
    "#006C00",
    "#FFFFFF",
    "#0000A6",
    "#FFFF00",
    "#FF0000",
    "#000000",
    "#000000",
]
cints = [-109.5, -109.0, -75.0, -47.0, -30.0, -15.5, 0.0, 0.5, 54.5]
cints = [(float(c + 109.5) / float(54.5 + 109.5)) for c in cints]
colorList = []
for i in list(range(0, len(cints))):
    colorList.append((cints[i], colors[i]))
cmap = mpcolors.LinearSegmentedColormap.from_list("mycmap", colorList)
for i in range(256):
    val = -109.5 + (i / 255) * (54.5 + 109.5)
    c = cmap(i / 255.0)
    print("%.2f %.0f %.0f %.0f" % (val, c[0] * 255, c[1] * 255, c[2] * 255))



"""Generic NWS Product Parser"""

# stdlib
from datetime import timedelta
from functools import partial

import click
from pyiem.database import get_sqlalchemy_conn
from pyiem.nws.product import TextProduct
from pyiem.nws.products import parser as productparser
from pyiem.nws.ugc import UGCProvider
from shapely.geometry import MultiPolygon

# Local
from pywwa import common
from pywwa.database import get_database, load_nwsli
from pywwa.ldm import bridge

NWSLI_DICT = {}


def process_data(ugc_dict, txn, buf) -> TextProduct:
    """Actually do some processing"""

    # Create our TextProduct instance
    prod = productparser(
        buf,
        utcnow=common.utcnow(),
        ugc_provider=ugc_dict,
        nwsli_provider=NWSLI_DICT,
    )

    # Do the Jabber work necessary after the database stuff has completed
    for plain, html, xtra in prod.get_jabbers(
        common.SETTINGS.get("pywwa_product_url", "pywwa_product_url")
    ):
        if xtra.get("channels", "") == "":
            common.email_error("xtra[channels] is empty!", buf)
        common.send_message(plain, html, xtra)

    if not common.dbwrite_enabled():
        return None
    # Insert into database only if there is a polygon!
    if not prod.segments or prod.segments[0].sbw is None:
        return None

    expire = prod.segments[0].ugcexpire
    if expire is None:
        prod.warnings.append("ugcexpire is none, defaulting to 90 minutes.")
        expire = prod.valid + timedelta(minutes=90)
    product_id = prod.get_product_id()
    giswkt = f"SRID=4326;{MultiPolygon([prod.segments[0].sbw]).wkt}"
    sql = (
        "INSERT into text_products(product_id, geom, issue, expire, pil) "
        "values (%s,%s,%s,%s,%s)"
    )
    myargs = (
        product_id,
        giswkt,
        prod.valid,
        expire,
        prod.afos,
    )
    txn.execute(sql, myargs)
    if prod.warnings:
        common.email_error("\n".join(prod.warnings), buf)
    return prod


@click.command(help=__doc__)
@common.init
def main(*args, **kwargs):
    """Go Main Go."""
    load_nwsli(NWSLI_DICT)
    with get_sqlalchemy_conn("postgis") as conn:
        ugc_dict = UGCProvider(pgconn=conn)
    func = partial(process_data, ugc_dict)
    bridge(func, dbpool=get_database("postgis"))



"""Process and Archive the NEXRAD Level III NCR Attribute Table"""

# stdlib
import math
from io import BytesIO
from zoneinfo import ZoneInfo

# 3rd Party
import click
from metpy.io.nexrad import Level3File
from twisted.internet import reactor

# Local
from pywwa import LOG, common
from pywwa.database import get_database
from pywwa.ldm import bridge

# Setup Database Links
PGCONN = get_database("radar", cp_max=5)

ST = {}


def load_station_table(txn):
    """Load the station table of NEXRAD sites"""
    LOG.info("load_station_table called() ...")
    txn.execute(
        "SELECT id, ST_x(geom) as lon, ST_y(geom) as lat from stations "
        "where network in ('NEXRAD','TWDR')"
    )
    for row in txn.fetchall():
        ST[row["id"]] = {"lat": row["lat"], "lon": row["lon"]}
    LOG.info("Station Table size %s", len(ST.keys()))


def process_data(data):
    """I am called when data is ahoy"""
    bio = BytesIO()
    bio.write(data)
    bio.seek(0)
    process(bio)


def process(bio):
    """Process our data, please"""
    l3 = Level3File(bio)
    ctx = {
        "nexrad": l3.siteID,
        "ts": l3.metadata["vol_time"].replace(tzinfo=ZoneInfo("UTC")),
        "lines": [],
    }
    if not hasattr(l3, "graph_pages"):
        LOG.info("%s %s has no graph_pages", ctx["nexrad"], ctx["ts"])
        return ctx
    for page in l3.graph_pages:
        for line in page:
            if "text" in line:
                ctx["lines"].append(line["text"])
    df = PGCONN.runInteraction(really_process, ctx)
    df.addErrback(common.email_error, ctx)
    return ctx


def delete_prev_attrs(txn, nexrad):
    """Remove any previous attributes for this nexrad"""
    txn.execute("DELETE from nexrad_attributes WHERE nexrad = %s", (nexrad,))


def really_process(txn, ctx):
    """
    This processes the output we get from the GEMPAK!

      STM ID  AZ/RAN TVS  MDA  POSH/POH/MX SIZE VIL DBZM  HT  TOP  FCST MVMT
         F0  108/ 76 NONE NONE    0/ 20/<0.50    16  53  8.3  16.2    NEW
         G0  115/ 88 NONE NONE    0/  0/ 0.00     8  47 10.2  19.3    NEW
         H0  322/ 98 NONE NONE    0/  0/ 0.00     6  43 12.1  22.2    NEW

      STM ID  AZ/RAN TVS  MDA  POSH/POH/MX SIZE VIL DBZM  HT  TOP  FCST MVMT
         H1  143/ 90 NONE NONE    0/  0/ 0.00     4  43 11.3  14.9  208/ 27
         Q5  125/ 66 NONE NONE    0/  0/ 0.00     3  42  9.3   9.3    NEW
         I8  154/ 73 NONE NONE    0/  0/ 0.00     1  33 10.9  10.9    NEW

         U8  154/126 NONE NONE     UNKNOWN       11  47 18.8  23.1  271/ 70
         J0  127/134 NONE NONE     UNKNOWN       24  51 20.2  33.9    NEW
    """
    delete_prev_attrs(txn, ctx["nexrad"])

    cenlat = float(ST[ctx["nexrad"]]["lat"])
    cenlon = float(ST[ctx["nexrad"]]["lon"])
    latscale = 111137.0
    lonscale = 111137.0 * math.cos(cenlat * math.pi / 180.0)

    #   STM ID  AZ/RAN TVS  MESO POSH/POH/MX SIZE VIL DBZM  HT  TOP  FCST MVMT
    co = 0
    for line in ctx["lines"]:
        if len(line) < 5:
            continue
        if line[1] != " ":
            continue
        tokens = line.replace(">", " ").replace("/", " ").split()
        if not tokens or tokens[0] == "STM":
            continue
        if tokens[5] == "UNKNOWN":
            tokens[5] = 0
            tokens.insert(5, 0)
            tokens.insert(5, 0)
        if len(tokens) < 13:
            LOG.info("Incomplete Line ||%s||", line)
            continue
        d = {}
        co += 1
        d["storm_id"] = tokens[0]
        d["azimuth"] = int(float(tokens[1]))
        if tokens[2] == "***":
            LOG.info("skipping bad line |%s|", line)
            continue
        d["range"] = int(float(tokens[2]) * 1.852)
        d["tvs"] = tokens[3]
        d["meso"] = tokens[4]
        d["posh"] = tokens[5] if tokens[5] != "***" else None
        d["poh"] = tokens[6] if tokens[6] != "***" else None
        if tokens[7] == "<0.50":
            tokens[7] = 0.01
        d["max_size"] = tokens[7]

        if tokens[8] in ["UNKNOWN", "***"]:
            d["vil"] = 0
        else:
            d["vil"] = tokens[8]

        d["max_dbz"] = tokens[9]
        d["max_dbz_height"] = tokens[10]
        d["top"] = tokens[11]
        if tokens[12] == "NEW":
            d["drct"] = 0
            d["sknt"] = 0
        else:
            d["drct"] = int(float(tokens[12]))
            d["sknt"] = tokens[13]
        d["nexrad"] = ctx["nexrad"]

        cosaz = math.cos(d["azimuth"] * math.pi / 180.0)
        sinaz = math.sin(d["azimuth"] * math.pi / 180.0)
        d["lat"] = cenlat + (cosaz * (d["range"] * 1000.0) / latscale)
        d["lon"] = cenlon + (sinaz * (d["range"] * 1000.0) / lonscale)
        d["valid"] = ctx["ts"]

        for table in [
            "nexrad_attributes",
            f"nexrad_attributes_{ctx['ts']:%Y}",
        ]:
            sql = f"""
                INSERT into {table} (nexrad, storm_id, geom, azimuth,
                range, tvs, meso, posh, poh, max_size, vil, max_dbz,
                max_dbz_height, top, drct, sknt, valid)
                values (%(nexrad)s, %(storm_id)s,
                ST_Point(%(lon)s, %(lat)s, 4326),
                %(azimuth)s, %(range)s, %(tvs)s, %(meso)s, %(posh)s,
                %(poh)s, %(max_size)s, %(vil)s, %(max_dbz)s,
                %(max_dbz_height)s, %(top)s, %(drct)s, %(sknt)s, %(valid)s)
            """
            if common.dbwrite_enabled():
                txn.execute(sql, d)

    if co > 0:
        LOG.info(
            "%s %s Processed %s entries",
            ctx["nexrad"],
            ctx["ts"].strftime("%Y-%m-%d %H:%M UTC"),
            co,
        )
    return co


def on_ready(_unused, mesosite):
    """ready to fire things up"""
    LOG.info("on_ready() has fired...")
    mesosite.close()
    bridge(process_data, isbinary=True)


def errback(res):
    """ERRORBACK"""
    LOG.error(res)
    reactor.stop()


@click.command(help=__doc__)
@common.init
@common.disable_xmpp
def main(*args, **kwargs):
    """Go Main Go"""
    mesosite = get_database("mesosite")
    df = mesosite.runInteraction(load_station_table)
    df.addCallback(on_ready, mesosite)
    df.addErrback(errback)



"""XTEUS Product Parser!"""

# 3rd Party
import click
from pyiem.nws.products.xteus import parser

# Local
from pywwa import common
from pywwa.database import get_database
from pywwa.ldm import bridge


def process_data(txn, data):
    """Process the product"""
    prod = parser(data, utcnow=common.utcnow())
    if prod.warnings:
        common.email_error("\n".join(prod.warnings), data)
    prod.sql(txn)
    return prod


@click.command(help=__doc__)
@common.init
@common.disable_xmpp
def main(*args, **kwargs):
    """Fire things up."""
    bridge(process_data, dbpool=get_database("iem"))



"""Produce a greys cmap."""

from pyiem.plot.use_agg import plt


def main():
    """Go Main Go."""
    cmap = plt.get_cmap("Greys_r")
    for i in range(256):
        c = cmap(i / 255.0)
        print("%s %.0f %.0f %.0f" % (i, c[0] * 255, c[1] * 255, c[2] * 255))


if __name__ == "__main__":
    main()



"""Parse CLI text products

The CLI report has lots of good data that is hard to find in other products,
so we take what data we find in this product and overwrite the database
storage of what we got from the automated observations
"""

from functools import partial

# 3rd Party
import click
from pyiem.network import Table as NetworkTable
from pyiem.nws.products import parser
from twisted.internet import reactor, task

# Local
from pywwa import LOG, common
from pywwa.database import get_database, get_dbconnc
from pywwa.ldm import bridge


def processor(nt, txn, text):
    """Protect the realprocessor"""
    prod = parser(
        text, nwsli_provider=nt.sts, utcnow=common.utcnow(), ugc_provider={}
    )
    # Run through database save now
    prod.sql(txn)
    jres = prod.get_jabbers(
        common.SETTINGS.get("pywwa_product_url", "pywwa_product_url")
    )
    for j in jres:
        common.send_message(j[0], j[1], j[2])
    if prod.warnings:
        common.email_error("\n".join(prod.warnings), text)
    return prod


def ready(nt):
    """We are ready."""
    func = partial(processor, nt)
    bridge(func, dbpool=get_database("iem"))


def load_stations():
    """load the stations and then fire the ingest."""
    # Can't use cursor= as it isn't an interator :(
    conn, cursor = get_dbconnc("mesosite")
    nt = NetworkTable("NWSCLI", only_online=False, cursor=cursor)
    LOG.info("Found %s NWSCLI stations", len(nt.sts))
    cursor.close()
    conn.close()
    return nt


@click.command(help=__doc__)
@common.init
def main(*args, **kwargs):
    """Go Main Go."""
    df = task.deferLater(reactor, 0, load_stations)
    df.addCallback(ready)
    df.addErrback(common.email_error)



"""Test afos_dump."""

# 3rd Party
import pytest
from pyiem.util import utc

# Local
import pywwa
from pywwa.testing import get_example_file
from pywwa.workflows import afos_dump


@pytest.mark.parametrize("database", ["afos"])
def test_future_product(cursor):
    """Test exception for product from the future."""
    data = get_example_file("AFD.txt")
    pywwa.CTX["utcnow"] = utc(2015, 6, 8, 11, 56)
    with pytest.raises(ValueError):
        afos_dump.process_data(cursor, data)


@pytest.mark.parametrize("database", ["afos"])
def test_memcache_write_api(cursor):
    """Exercise, but does not actually run the write :("""
    data = get_example_file("AFD.txt")
    pywwa.CTX["utcnow"] = utc(2015, 6, 9, 11, 56)
    prod = afos_dump.process_data(cursor, data)
    afos_dump.write2memcache(prod.get_product_id(), prod.unixtext)


@pytest.mark.parametrize("database", ["afos"])
def test_processor(cursor):
    """Test basic parsing."""
    data = get_example_file("AFD.txt")
    # 0. Very latent product
    with pytest.raises(ValueError):
        afos_dump.process_data(cursor, data)
    pywwa.CTX["utcnow"] = utc(2015, 6, 9, 11, 56)
    # 1. straight through
    prod = afos_dump.process_data(cursor, data)
    assert prod.afos == "AFDDMX"
    # 2. Corrupt the AFOS
    data = data.replace("AFDDMX", "AFDDMX123")
    with pytest.raises(ValueError):
        afos_dump.process_data(cursor, data)
    # 3. Corrupt the AFOS and WMO source
    data = data.replace("KDMX", "QQQQ")
    res = afos_dump.process_data(cursor, data)
    assert res is None
    # 4. Have something in exclude from memcache
    data = data.replace("AFDDMX123", "RR1DMX")
    res = afos_dump.process_data(cursor, data)
    assert res is None
    # 6. Replace on
    pywwa.CTX["replace"] = True
    afos_dump.process_data(cursor, data)


def test_empty():
    """Test what happens when we pass an emptry string."""
    assert afos_dump.process_data(None, "") is None


@pytest.mark.parametrize("database", ["afos"])
def test_process_data(cursor):
    """Test API for now, noop."""
    afos_dump.process_data(cursor, "")



"""Test bufr_surface workflow."""

import os

import pytest

from pywwa.database import get_dbconnc
from pywwa.testing import get_example_filepath
from pywwa.workflows import bufr_surface

# Hmmmm
bufr_surface.load_wmo2iso3166()


def sync_workflow(cursor, buffn) -> int:
    """Run twice through workflow, eh"""
    mconn, mcursor = get_dbconnc("mesosite")
    bufr_surface.load_xref(mcursor)
    with open(get_example_filepath(f"BUFR/{buffn}"), "rb") as fh:
        prod, datalists = bufr_surface.workflow(fh.read(), cursor, mcursor)
    mconn.close()
    obs = []
    for datalist in datalists:
        obs.append(bufr_surface.datalist2iemob_data(datalist, prod.source))
    return prod, obs


def generate_testfiles():
    """Generate a list of test files in the examples/BUFR directory."""
    for filename in os.listdir(get_example_filepath("BUFR")):
        if not filename.endswith(".bufr"):
            continue
        yield filename


def test_phour():
    """Test that we get a message with a phour set."""
    prod, obs = sync_workflow(None, "ISAI01_SBBR.bufr")
    assert prod.source == "SBBR"
    assert abs(obs[182]["phour"] - 0.0079) < 0.0001


@pytest.mark.parametrize("database", ["iem"])
def test_240103_bad_date(cursor):
    """Test failure in the wild."""
    sync_workflow(cursor, "ISCI01_SABM.bufr")


@pytest.mark.parametrize("database", ["iem"])
def test_231208_null_bytes(cursor):
    """Test failure in the wild."""
    sync_workflow(cursor, "ISMA01_FNLU_badid2.bufr")


@pytest.mark.parametrize("database", ["iem"])
@pytest.mark.parametrize("buffn", generate_testfiles())
def test_bufr_files_in_examples(cursor, buffn):
    """Parse all our examples."""
    sync_workflow(cursor, buffn)


def test_bounds_check():
    """Test that we can handle bad bounds."""
    assert bufr_surface.bounds_check(10, 0, 3) is None
    assert bufr_surface.bounds_check(None, 0, 10) is None


@pytest.mark.parametrize("database", ["iem"])
def test_unknown_source(cursor):
    """Test that invalid descriptor is handled."""
    with open(get_example_filepath("BUFR/ISIA14_CWAO.bufr"), "rb") as fh:
        payload = fh.read()
    bufr_surface.workflow(payload.replace(b"CWAO", b"XXXX"))


@pytest.mark.parametrize("database", ["mesosite"])
def test_load_xref(cursor):
    """Test loading the xref."""
    cursor.execute(
        "insert into stations(id, wigos, network, iemid) "
        "values ('SCSC', 'SCSC', %s, -1)",
        (bufr_surface.NETWORK,),
    )
    bufr_surface.load_xref(cursor)



"""Generate some CONUS composites like we had back in the day.

1. gdalwarp: convert East and West into CONUS EPSG:4326 at desired res ~4km
   appears CONUS goes west goes to just -112 east?
2. gdal_merge.py: combine these two with the west file last (on top)
"""

import datetime
import json
import os
import subprocess
import sys
import tempfile
from zoneinfo import ZoneInfo

from pyiem.util import logger, utc

ISO = "%Y-%m-%dT%H:%M:%SZ"
LOG = logger()
PATH = "/mesonet/ldmdata/gis/images/GOES/conus"
LOOKUP = {2: "vis", 9: "wv", 13: "ir"}


def filtermsg(buf):
    """Less is more"""
    text = buf.decode("ascii")
    return "Point outside of projection" in text


def call(cmd: list):
    """Wrapper around subprocess."""
    strcmd = " ".join(cmd)
    LOG.info(strcmd)
    proc = subprocess.Popen(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    stdout = proc.stdout.read()
    stderr = proc.stderr.read()
    if stdout != b"":
        if not filtermsg(stdout):
            LOG.warning(strcmd)
            LOG.warning(stdout)
    if stderr != b"":
        if not filtermsg(stderr):
            LOG.warning(strcmd)
            LOG.warning(stderr)


def run(valid, channel, tmpname):
    """Do work."""
    tiles = []
    for bird in [16, 17, 18, 19]:
        fnbase = f"{PATH}/channel{channel:02.0f}/GOES-{bird}_C{channel:02.0f}"
        if not os.path.isfile(f"{fnbase}.json"):
            LOG.info("Missing %s.json", fnbase)
            continue
        with open(f"{fnbase}.json", encoding="utf8") as fp:
            info = json.load(fp)
        ts = datetime.datetime.strptime(info["meta"]["valid"], ISO).replace(
            tzinfo=ZoneInfo("UTC")
        )
        if (valid - ts) > datetime.timedelta(hours=6):
            LOG.info("Skipping %s, too old", fnbase)
            continue
        # step 1
        fn = f"{tmpname}_{bird}.tif"
        tiles.append(fn)
        cmd = [
            "gdalwarp",
            "-q",
            "-s_srs",
            info["meta"]["proj4str"],
            "-t_srs",
            "EPSG:4326",
            "-te",
            "-140" if bird in [17, 18] else "-113.1",
            "22",
            "-112.9" if bird in [17, 18] else "-60",
            "52",
            "-tr",
            "0.04",
            "0.04",
            "-of",
            "GTiff",
            f"{fnbase}.png",
            fn,
        ]
        call(cmd)

    if not tiles:
        LOG.warning("No data found for tiling...")
        return
    cmd = [
        "gdal_merge.py",
        "-q",
        "-o",
        f"{tmpname}.tif",
        "-of",
        "GTiff",
        "-ps",
        "0.04",
        "0.04",
        "-ul_lr",
        "-130",
        "52",
        "-60",
        "22",
        *tiles,
    ]
    call(cmd)

    # GIS/sat/conus_goes_vis4km_1045.tif
    cmd = [
        "pqinsert",
        "-i",
        "-p",
        (
            f"gis ac {valid:%Y%m%d%H%M} "
            f"gis/images/4326/sat/conus_goes_{LOOKUP[channel]}4km.tif "
            f"GIS/sat/conus_goes_{LOOKUP[channel]}4km_{valid:%H%M}.tif tif"
        ),
        f"{tmpname}.tif",
    ]
    call(cmd)


def main(argv):
    """Go Main Go."""
    valid = utc(*[int(a) for a in argv[1:6]])
    LOG.info("valid is set to: %s", valid)
    for channel in [2, 9, 13]:
        with tempfile.NamedTemporaryFile() as tmpfd:
            try:
                run(valid, channel, tmpfd.name)
            except Exception as exp:
                LOG.exception(exp)
        for part in ["", "_16", "_17", "_18", "_19"]:
            fn = f"{tmpfd.name}{part}.tif"
            if not os.path.isfile(fn):
                continue
            os.unlink(fn)


if __name__ == "__main__":
    main(sys.argv)



"""
VTEC product ingestor

The warnings table has the following timestamp based columns, this gets ugly
with watches.  Lets try to explain

    issue   <- VTEC timestamp of when this event was valid for
    expire  <- When does this VTEC product expire
    updated <- Product Timestamp of when a product gets updated
    init_expire <- When did this product initially expire
    product_issue <- When was this product issued by the NWS
"""

from functools import partial

# 3rd Party
import click
from pyiem.database import get_sqlalchemy_conn
from pyiem.nws.products.vtec import parser as vtecparser
from pyiem.nws.ugc import UGCProvider
from twisted.mail.smtp import SMTPSenderFactory

# Local
from pywwa import LOG, common
from pywwa.database import get_database, load_nwsli
from pywwa.ldm import bridge

SMTPSenderFactory.noisy = False
NWSLI_DICT = {}


def process_data(ugc_dict, pgconn, data):
    """Process the product"""
    # Make sure we have a trailing $$, if not report error and slap one on
    if data.find("$$") == -1:
        common.email_error("No $$ Found!", data)
        data += "\n\n$$\n\n"

    # Create our TextProduct instance
    text_product = vtecparser(
        data,
        utcnow=common.utcnow(),
        ugc_provider=ugc_dict,
        nwsli_provider=NWSLI_DICT,
    )
    # Don't parse these as they contain duplicated information
    if text_product.source == "KNHC" and text_product.afos[:3] == "TCV":
        return
    # Skip spanish products
    if text_product.source == "TJSJ" and text_product.afos[3:] == "SPN":
        return

    # This is sort of ambiguous as to what is best to be done when database
    # writing is disabled.  The web workflow will likely fail as well.
    if common.dbwrite_enabled():
        df = pgconn.runInteraction(text_product.sql)
        df.addCallback(step2, text_product)
        df.addErrback(common.email_error, text_product.unixtext)
        df.addErrback(LOG.error)
    else:
        step2(None, text_product)


def step2(_dummy, text_product):
    """Callback after hopeful database work."""
    if text_product.warnings:
        common.email_error(
            "\n\n".join(text_product.warnings), text_product.text
        )

    # Do the Jabber work necessary after the database stuff has completed
    for plain, html, xtra in text_product.get_jabbers(
        common.SETTINGS.get("pywwa_vtec_url", "pywwa_vtec_url"),
        common.SETTINGS.get("pywwa_river_url", "pywwa_river_url"),
    ):
        if xtra.get("channels", "") == "":
            common.email_error("xtra[channels] is empty!", text_product.text)
        common.send_message(plain, html, xtra)


@click.command(help=__doc__)
@common.init
def main(*args, **kwargs):
    """Go Main Go."""
    load_nwsli(NWSLI_DICT)
    with get_sqlalchemy_conn("postgis") as conn:
        ugc_dict = UGCProvider(pgconn=conn)
    bridge(partial(process_data, ugc_dict, get_database("postgis")))



"""Test dsm2afos workflow."""

from pywwa.testing import get_example_file
from pywwa.workflows.dsm2afos import workflow


def test_workflow():
    """run workflow."""
    raw = get_example_file("DSM.txt")
    workflow(raw)



"""test mos workflow"""

import pytest

from pywwa.testing import get_example_file
from pywwa.workflows.mos import process_data


@pytest.mark.parametrize("database", ["mos"])
def test_workflow(cursor):
    """Test."""
    process_data(cursor, get_example_file("METNC1.txt"))



"""An infinite loop to get ridge_runner.py going."""

import os
import subprocess
import time


def main():
    """Infinite loop here"""
    scriptfn = os.sep.join([os.path.dirname(__file__), "ridge_processor.py"])
    while True:
        print("ridge_runner.py is execing ridge_processor.py now...")
        subprocess.call(["python", scriptfn])
        time.sleep(3)


if __name__ == "__main__":
    main()



"""Repair holes when aviationweather.gov METAR service has data."""

import subprocess
import sys
import time

import click
import httpx
import pandas as pd
from pyiem.database import get_sqlalchemy_conn, sql_helper
from pyiem.util import logger, utc
from tqdm import tqdm

LOG = logger()
FN = "/mesonet/tmp/AWX_METARS.txt"


@click.command()
@click.option("--network", help="Network to query, defaults to all.")
def main(network):
    """Go Main Go."""
    nlim = "network ~* 'ASOS'" if network is None else "network = :network"
    with get_sqlalchemy_conn("mesosite") as conn:
        # don't do online check as awx may have data we don't
        stations = pd.read_sql(
            sql_helper(
                "select iemid, id, network from stations where {nlim} "
                "order by id ASC",
                nlim=nlim,
            ),
            conn,
            params={"network": network},
            index_col="iemid",
        )
    stations["found"] = 0
    with (
        open(FN, "wb") as fh,
        httpx.Client() as client,
        get_sqlalchemy_conn("iem") as conn,
    ):
        # create faked noaaport header
        fh.write(b"000 \r\r\n")
        fh.write(f"SAUS70 KISU {utc():%d%H%M}\r\r\n".encode("ascii"))
        progress = tqdm(
            stations.iterrows(),
            disable=not sys.stdout.isatty(),
            total=len(stations.index),
        )
        for iemid, row in progress:
            st4 = row["id"] if len(row["id"]) == 4 else f"K{row['id']}"
            url = (
                "https://aviationweather.gov/api/data/metar?"
                f"ids={st4}&format=raw&hours=36"
            )
            attempt = 0
            while attempt < 3:
                attempt += 1
                try:
                    resp = client.get(url, timeout=20)
                    if resp.status_code == 204:
                        if sys.stdout.isatty():
                            progress.write(f"No data for {st4}")
                        break
                    # 429 is a backoff warning
                    # 502 is a broken proxy service
                    if resp.status_code in [429, 502]:
                        LOG.info(
                            "Got %s, cooling jets for 5 seconds.",
                            resp.status_code,
                        )
                        time.sleep(5)
                        continue
                    if resp.status_code != 200:
                        LOG.warning("Failure %s %s", st4, resp.status_code)
                        continue
                    break
                except Exception as exp:
                    LOG.info("Failed to fetch %s: %s", st4, exp)
            awx = {}
            for line in resp.text.split("\n"):
                if line.strip() == "":
                    continue
                awx[line[5:11]] = f"{line}="
            res = conn.execute(
                sql_helper(
                    "select valid at time zone 'UTC' as v from current_log "
                    "WHERE iemid = :iemid and "
                    "valid > now() - '50 hours'::interval "
                    "and raw !~* 'MADISHF' "
                    "ORDER by valid ASC"
                ),
                {"iemid": iemid},
            )

            for row2 in res:
                key = f"{row2[0]:%d%H%M}"
                if key in awx:
                    awx.pop(key)
            if awx:
                stations.at[iemid, "found"] = len(awx)
                for _, val in awx.items():
                    fh.write(f"{val}\r\r\n".encode("ascii"))
            # throttle requests to something around 5/s
            time.sleep(0.2)
        fh.write(b"\003")

    LOG.info("Found %s METARs", stations["found"].sum())
    with subprocess.Popen(
        [
            "pywwa-parse-metar",
            "-x",
            "-s",
            "120",
        ],
        stdin=subprocess.PIPE,
    ) as proc:
        with open(FN, "rb") as fh:
            proc.stdin.write(fh.read())
        proc.stdin.close()
        proc.wait()

    # Insert into LDM for archival
    subprocess.call(
        [
            "pqinsert",
            "-p",
            f"data a {utc():%Y%m%d%H%M} text/supp_metars_via_avwxgov.txt "
            "text/supp_metars_via_avwxgov.txt txt",
            FN,
        ],
    )


if __name__ == "__main__":
    main()



"""Test (legacy) ERO parsing."""

import pytest

from pywwa.testing import get_example_file
from pywwa.workflows import ero


@pytest.mark.parametrize("database", ["postgis"])
def test_basic(cursor):
    """Test goodline"""
    prod = ero.real_parser(cursor, get_example_file("RBG94E.txt"))
    ero.do_jabber(prod)
    assert prod is not None



"""SPC Watch (SAW, SEL, WWP) Ingestor"""

# 3rd Party
import click
from pyiem.nws.product import TextProduct
from pyiem.nws.products import parser
from twisted.internet.task import LoopingCall

# Local
from pywwa import LOG, common
from pywwa.database import get_database
from pywwa.ldm import bridge

# Create a running queue of watch products, so that once we get the
# collective, we can generate the jabbers and send them out.
QUEUE = {}


def do_jabber(entry):
    """We are ready now, send out the jabbers"""
    baseurl = common.SETTINGS.get("pywwa_watch_url", "pywwa_watch_url")
    prod = entry["SAW"]
    jmsgs = prod.get_jabbers(
        baseurl,
        wwpprod=entry["WWP"],
        selprod=entry["SEL"],
    )
    for txt, html, xtra in jmsgs:
        common.send_message(txt, html, xtra)


def process_queue():
    """Process the queue every minute to see if we have complete messages."""
    for wnum in list(QUEUE.keys()):
        entry = QUEUE[wnum]
        QUEUE[wnum]["loops"] += 1
        # 15 seconds, we await 4 minutes
        if QUEUE[wnum]["loops"] > 16:
            if entry["SAW"] is not None:
                do_jabber(entry)
            else:
                common.email_error(f"SAW Missing for {wnum}", repr(entry))
            QUEUE.pop(wnum)
            continue
        if None in [entry["SAW"], entry["SEL"], entry["WWP"]]:
            LOG.warning(
                "Watch %s[SAW:%s,SEL:%s,WWP:%s] is not complete, waiting",
                wnum,
                "miss" if entry["SAW"] is None else "X",
                "miss" if entry["SEL"] is None else "X",
                "miss" if entry["WWP"] is None else "X",
            )
            continue
        do_jabber(entry)
        QUEUE.pop(wnum)


def real_process(txn, raw) -> TextProduct:
    """Process the product, please"""
    prod = parser(raw, utcnow=common.utcnow(), ugc_provider={})
    LOG.info("Watch %s received", prod.get_product_id())
    # NOTE: ensure parsers are implmenting the same interface
    if prod.is_test():
        LOG.info("TEST watch found %s, skipping", prod.get_product_id())
        return prod
    if common.dbwrite_enabled():
        prod.sql(txn)
    if prod.warnings:
        common.email_error("\n".join(prod.warnings), raw)

    if prod.afos[:3] != "SAW":
        wnum = prod.data.num
    else:
        prod.compute_wfos(txn)
        wnum = prod.ww_num

    res = QUEUE.setdefault(
        wnum, {"SAW": None, "SEL": None, "WWP": None, "loops": 0}
    )
    res[prod.afos[:3]] = prod
    return prod


@click.command(help=__doc__)
@common.init
def main(*args, **kwargs):
    """Go Main Go"""
    bridge(real_process, dbpool=get_database("postgis"))
    lc = LoopingCall(process_queue)
    df = lc.start(15, now=False)
    df.addErrback(common.email_error)



"""Test aviation."""

# 3rd Party
import pytest
from pyiem.database import get_dbconnc
from pyiem.util import utc

# Local
import pywwa
from pywwa.testing import get_example_file
from pywwa.workflows import aviation


@pytest.mark.parametrize("database", ["postgis"])
def test_processor(cursor):
    """Test basic parsing."""
    mconn, mcursor = get_dbconnc("mesosite")
    data = get_example_file("SIGC.txt")
    pywwa.CTX["utcnow"] = utc(2005, 11, 13, 6, 55)
    aviation.load_database(mcursor)
    mconn.close()
    prod = aviation.process_data(cursor, data)
    # 1. Create a bad loc
    aviation.process_data(cursor, data.replace(" ORL", " ZZZ"))
    # 2. Attempt Jabber
    aviation.final_step(prod)


@pytest.mark.parametrize("database", ["mesosite"])
def test_load_database(cursor):
    """Test our load_database function."""
    aviation.load_database(cursor)
    assert isinstance(aviation.LOCS, dict)



"""TAF Ingestor"""

import click
from pyiem.nws.products.taf import parser

from pywwa import common
from pywwa.database import get_database
from pywwa.ldm import bridge


def real_process(txn, raw):
    """Process the product, please"""
    prod = parser(raw, utcnow=common.utcnow(), ugc_provider={})
    if common.dbwrite_enabled():
        prod.sql(txn)
    baseurl = common.SETTINGS.get("pywwa_product_url", "pywwa_product_url")
    jmsgs = prod.get_jabbers(baseurl)
    for mess, htmlmess, xtra in jmsgs:
        common.send_message(mess, htmlmess, xtra)
    if prod.warnings:
        common.email_error("\n\n".join(prod.warnings), prod.text)


@click.command(help=__doc__)
@common.init
def main(*args, **kwargs):
    """Go Main Go"""
    bridge(real_process, dbpool=get_database("asos"))



"""Archive MADIS files."""

import glob
import os
import subprocess
from datetime import date, timedelta

from pyiem.util import logger

LOG = logger()
HOST = "akrherz-desktop.agron.iastate.edu"


def process(mydir, valid):
    """Do the necessary work."""
    fulldir = f"/mesonet/data/madis/{mydir}"
    if not os.path.isdir(fulldir):
        LOG.info("Directory %s does not exist", fulldir)
        return
    os.chdir(fulldir)
    # Collapse files down.
    for hr in range(0, 24):
        i = 300
        found = False
        while i >= 0:
            fn = f"{valid:%Y%m%d}_{hr:02.0f}00_{i}.nc"
            if os.path.isfile(fn):
                if not found:
                    found = True
                    cmd = ["cp", "-f", fn, f"{valid:%Y%m%d}_{hr:02.0f}00.nc"]
                    subprocess.call(cmd)
                os.unlink(fn)
            i -= 1
    # gzip
    files = glob.glob(f"{valid:%Y%m%d}*nc")
    if not files:
        LOG.info("No files found for %s", mydir)
        return
    subprocess.call(["gzip", *files])
    files = glob.glob(f"{valid:%Y%m%d}*nc.gz")
    if not files:
        LOG.info("No gzip files found for %s", mydir)
        return
    # rsync
    cmd = [
        "rsync",
        "-a",
        "--remove-source-files",
        "--rsync-path",
        f"mkdir -p /stage/iemoffline/madis/{mydir}/{valid:%Y} && rsync",
        *files,
        f"{HOST}:/stage/iemoffline/madis/{mydir}/{valid:%Y}/",
    ]
    subprocess.call(cmd)


def main():
    """Go Main Go."""
    valid = date.today() - timedelta(days=1)
    for mydir in ["hfmetar", "mesonet1", "metar", "rwis1"]:
        process(mydir, valid)


if __name__ == "__main__":
    main()



"""
Handle things that need emailed to me for my situational awareness.
"""

# stdlib
import re
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

import click
from pyiem.nws import product
from twisted.mail import smtp

# Local
from pywwa import LOG, common
from pywwa.ldm import bridge

IOWA_WFOS = ["KDMX", "KDVN", "KARX", "KFSD", "KOAX"]
EF_RE = re.compile(r"^Rating:\s*EF\s?\-?(?P<num>\d)\s*$", re.M | re.I)


def process_data(data):
    """
    Actual ingestor
    """
    try:
        real_process(data)
    except Exception as exp:
        common.email_error(exp, data)


def damage_survey_pns(prod):
    """Glean out things, hopefully."""
    subject = f"Damage Survey PNS from {prod.source}"
    plain = prod.unixtext
    ffs = {}
    url = (
        f"https://mesonet.agron.iastate.edu/p.php?pid={prod.get_product_id()}"
    )
    maxtext = (
        f'<hr><p>Text Permalink: <a href="{url}">IEM Website</a> '
        f"({prod.get_product_id()})</p>\n<hr>"
    )
    for token in EF_RE.findall(plain):
        entry = ffs.setdefault(int(token), [])
        entry.append(1)
    if ffs:
        maxf = max(ffs.keys())
        table = ""
        for ef in range(6):
            if ef not in ffs:
                continue
            table += f"EF-{ef} ⇒ {len(ffs[ef])}<br />\n"
        subject = f"Damage Survey PNS (Max: EF{maxf}) from {prod.source}"
        maxtext += (
            f"<p>Max EF Rating Below: <strong>(EF{maxf})</strong></p>"
            "<p>Count by Rating:</p>"
            "<p><pre>"
            f"{table}"
            "</pre></p>"
        )
    html = f"{maxtext}<hr><p><pre>{plain}</pre></p>"
    return (
        subject,
        MIMEText(plain, "plain", "utf-8"),
        MIMEText(html, "html", "utf-8"),
    )


def real_process(data) -> product.TextProduct:
    """Go!"""
    prod = product.TextProduct(data, ugc_provider={})
    if prod.afos == "ADMNES":
        LOG.warning("Dumping %s on the floor", prod.get_product_id())
        return None

    # Strip off stuff at the top
    msg = MIMEMultipart("alternative")
    msgtext = MIMEText(prod.unixtext, "plain", "utf-8")
    msghtml = MIMEText(f"<p><pre>{prod.unixtext}</pre></p>", "html", "utf-8")
    # some products have no AWIPS ID, sigh
    subject = prod.wmo
    msg["To"] = "akrherz@iastate.edu"
    msg["From"] = common.SETTINGS.get("pywwa_errors_from", "ldm@localhost")
    cc = None
    if prod.afos is not None:
        subject = prod.afos
        if prod.afos[:3] == "ADM":
            subject = f"ADMIN NOTICE {prod.afos[3:]}"
        elif prod.afos[:3] == "PNS" and prod.afos != "PNSWSH":
            if prod.unixtext.upper().find("DAMAGE SURVEY") == -1:
                return None
            subject = f"Damage Survey PNS from {prod.source}"
            try:
                subject, msgtext, msghtml = damage_survey_pns(prod)
            except Exception as exp:
                LOG.error(exp)
                common.email_error(exp, prod.unixtext)
            del msg["To"]
            msg["To"] = "nws-damage-survey-pns@googlegroups.com"
            del msg["From"]
            msg["From"] = "akrherz@iastate.edu"
        elif prod.afos[:3] == "RER":
            subject = f"[RER] {prod.source} {prod.afos[3:]}"
            if prod.source in IOWA_WFOS:
                cc = "Justin.Glisan@iowaagriculture.gov"
    addrs = [
        msg["To"],
    ]
    if cc is not None:
        msg["Cc"] = cc
        addrs.append(cc)
    msg["subject"] = subject
    msg.attach(msgtext)
    msg.attach(msghtml)
    df = smtp.sendmail(
        common.SETTINGS.get("pywwa_smtp", "smtp"), msg["From"], addrs, msg
    )
    df.addErrback(LOG.error)
    return prod


@click.command(help=__doc__)
@common.init
@common.disable_xmpp
def main(*args, **kwargs):
    """Go Main Go."""
    bridge(process_data)



"""Test cli_parser."""

# 3rd Party
import pytest
from pyiem.network import Table as NetworkTable
from pyiem.util import utc

# Local
import pywwa
from pywwa.testing import get_example_file
from pywwa.workflows import cli


@pytest.mark.parametrize("database", ["iem"])
def test_processor(cursor):
    """Test basic parsing."""
    data = get_example_file("CLI.txt")
    pywwa.CTX["utcnow"] = utc(2015, 6, 9, 6, 51)
    nt = NetworkTable("NWSCLI", only_online=False)
    prod = cli.processor(nt, cursor, data)
    assert prod.valid == pywwa.CTX["utcnow"]


@pytest.mark.parametrize("database", ["iem"])
def test_two_clis(cursor):
    """Test parsing ye infamous double CLI."""
    cursor.execute(
        "select max_tmpf from summary_2014 s JOIN stations t "
        "on (s.iemid = t.iemid) WHERE t.id in ('HOU', 'IAH') "
        "and day = '2014-11-30'"
    )
    data = get_example_file("CLIHOU.txt")
    nt = NetworkTable("NWSCLI", only_online=False)
    prod = cli.processor(nt, cursor, data)
    assert len(prod.data) == 2


@pytest.mark.parametrize("database", ["iem"])
def test_bad_station(cursor):
    """Test what happens when we have an unknown station."""
    data = get_example_file("CLI.txt").replace("CLIFGF", "CLIXXX")
    pywwa.CTX["utcnow"] = utc(2015, 6, 9, 6, 51)
    nt = NetworkTable("NWSCLI", only_online=False)
    prod = cli.processor(nt, cursor, data)
    assert prod is not None



"""
My purpose in life is to take the NWS AWIPS Geodata Zones Shapefile and
dump them into the PostGIS database!  I was bootstraped like so:

python ugcs_update.py z_16mr06 2006 03 16
python ugcs_update.py z_11mr07 2007 03 11
python ugcs_update.py z_31my07 2007 05 31
python ugcs_update.py z_01au07 2007 08 01
python ugcs_update.py z_5sep07 2007 09 05
python ugcs_update.py z_25sep07 2007 09 25
python ugcs_update.py z_01ap08 2008 04 01
python ugcs_update.py z_09se08 2008 09 09
python ugcs_update.py z_03oc08 2008 10 03
python ugcs_update.py z_07my09 2009 05 07
python ugcs_update.py z_15jl09 2009 07 15
python ugcs_update.py z_22jl09 2009 07 22
python ugcs_update.py z_04au11 2011 08 04
python ugcs_update.py z_13oc11 2011 10 13
python ugcs_update.py z_31my11 2011 05 31
python ugcs_update.py z_15de11 2011 12 15
python ugcs_update.py z_23fe12 2012 02 23
python ugcs_update.py z_03ap12 2012 04 03
python ugcs_update.py z_12ap12 2012 04 12
python ugcs_update.py z_07jn12 2012 06 07
python ugcs_update.py z_11oc12 2012 10 11
python ugcs_update.py z_03de13a 2013 12 03
python ugcs_update.py z_05fe14a 2014 02 05

"""

import os
import sys
import zipfile
from datetime import datetime, timezone

import click
import geopandas as gpd
import httpx
import pandas as pd
from psycopg.rows import dict_row
from pyiem.database import get_sqlalchemy_conn, sql_helper
from pyiem.util import logger
from shapely.geometry import MultiPolygon

# Some strangely encoded WFOs need to be rectified
WFO_XREF = {
    "PQW": "GUM",
    "PQE": "GUM",
}

LOG = logger()
# Change Directory to /tmp, so that we can rw
os.chdir("/tmp")


def do_download(zipfn):
    """Do the download steps"""
    if not os.path.isfile(zipfn):
        url = (
            "https://www.weather.gov/source/gis/Shapefiles/"
            f"{'County' if zipfn.startswith('c_') else 'WSOM'}/{zipfn}"
        )
        req = httpx.get(url, timeout=60)
        LOG.info("Downloading %s ...", url)
        if req.status_code != 200:
            LOG.warning("Download %s failed, got %s", zipfn, req.status_code)
            sys.exit(1)
        with open(zipfn, "wb") as fh:
            fh.write(req.content)

    LOG.info("Unzipping")
    shpfn = None
    with zipfile.ZipFile(zipfn, "r") as zipfp:
        for name in zipfp.namelist():
            LOG.info("Extracting %s", name)
            with open(name, "wb") as fh:
                fh.write(zipfp.read(name))
            if name[-3:] == "shp":
                shpfn = name
    return shpfn


def new_poly(geo):
    """Sort and return new multipolygon"""
    if geo.geom_type == "Polygon":
        return geo
    # This is tricky. We want to have our multipolygon have its
    # biggest polygon first in the multipolygon.
    # This will allow some plotting simplification
    # later as we will only consider the first polygon
    maxarea = 0
    polys = []
    for poly in geo.geoms:
        area = poly.area
        if area > maxarea:
            maxarea = area
            polys.insert(0, poly)
        else:
            polys.append(poly)

    return MultiPolygon(polys)


def db_fixes(cursor, valid):
    """Fix some issues in the database"""
    for year in [2000, 2005, 2010, 2015, 2020]:
        table = f"gpw{year}"
        col = f"gpw_population_{year}"
        res = cursor.execute(
            sql_helper(
                """
    with pops as (
        select gid, coalesce(sum(population), 0) as pop
        from ugcs u LEFT JOIN {table} g on st_contains(u.geom, g.geom) where
        u.{col} is null GROUP by gid)
    update ugcs u SET {col} = pop FROM pops p WHERE u.gid = p.gid""",
                table=table,
                col=col,
            )
        )
        LOG.info("Set %s rows for gpw_population_%s", res.rowcount, year)

    res = cursor.execute(
        sql_helper("""
    update ugcs SET geom = st_makevalid(geom) where end_ts is null
    and not st_isvalid(geom) and begin_ts = :begin"""),
        {"begin": valid},
    )
    LOG.info("Fixed %s entries that were ST_Invalid()", res.rowcount)

    res = cursor.execute(
        sql_helper("""
        UPDATE ugcs SET simple_geom = ST_Multi(
            ST_Buffer(ST_SnapToGrid(geom, 0.01), 0)
        ),
        centroid = ST_Centroid(geom),
        area2163 = ST_area( ST_transform(geom, 2163) ) / 1000000.0
        WHERE begin_ts = :valid or area2163 is null
    """),
        {
            "valid": valid,
        },
    )
    LOG.info("Updated simple_geom,centroid,area2163 for %s rows", res.rowcount)

    # Check the last step that we don't have empty geoms, which happened once
    def _check():
        """Do the check."""
        return cursor.execute(
            sql_helper("""
            SELECT end_ts from ugcs
            where begin_ts = :valid and (
                simple_geom is null or
                ST_IsEmpty(simple_geom) or
                ST_Area(simple_geom) / ST_Area(geom) < 0.9
            )
        """),
            {
                "valid": valid,
            },
        )

    res = _check()
    if res.rowcount > 0:
        LOG.info(
            "%s rows with empty, too small simple_geom, decreasing tolerance",
            res.rowcount,
        )
        cursor.execute(
            sql_helper("""
            UPDATE ugcs
            SET simple_geom = ST_Multi(
                ST_Buffer(ST_SnapToGrid(geom, 0.0001), 0)
            )
            WHERE begin_ts = :valid and (
                simple_geom is null or
                ST_IsEmpty(simple_geom) or
                ST_Area(simple_geom) / ST_Area(geom) < 0.9
            )
        """),
            {
                "valid": valid,
            },
        )
        res = _check()
        if res.rowcount > 0:
            LOG.info(
                "Found %s rows with empty simple_geom, FIXME SOMEHOW!",
                res.rowcount,
            )


def truncate(cursor, valid, ugc, source, ctid):
    """Stop the bleeding."""
    res = cursor.execute(
        sql_helper("""
    UPDATE ugcs SET end_ts = :valid WHERE ugc = :ugc and end_ts is null
    and source = :source and ctid = :ctid"""),
        {
            "valid": valid,
            "ugc": ugc,
            "source": source,
            "ctid": ctid,
        },
    )
    return res.rowcount


def read_shapefile(shpfn: str) -> pd.DataFrame:
    """Standardized stuff."""
    df = gpd.read_file(shpfn)
    # Ensure CRS is set
    df["geometry"] = df["geometry"].set_crs("EPSG:4326", allow_override=True)
    if df.empty:
        LOG.info("Abort, empty dataframe from shapefile read.")
        sys.exit()
    # make all columns upper
    df.columns = [x.upper() if x != "geometry" else x for x in df.columns]
    return df


def chunk_df(df: pd.DataFrame):
    for ugc, gdfin in df.groupby("ugc"):
        gdf = gdfin.reset_index()
        if len(gdf.index) == 1:
            yield gdf.iloc[0].to_dict()
            continue
        LOG.info("Found %s rows for %s", len(gdf.index), ugc)
        # If name and cwa are the same, we can just merge the polygons
        if gdf["NAME"].nunique() == 1 and gdf["wfo"].nunique() == 1:
            LOG.info("--> Merging geometry of %s", ugc)
            yield {
                "ugc": ugc,
                "geometry": new_poly(gdf.union_all()),  # type: ignore
                "area2163": gdf["area2163"].sum(),
                "NAME": gdf.iloc[0]["NAME"],
                "STATE": gdf.iloc[0]["STATE"],
                "wfo": gdf.iloc[0]["wfo"],
            }
            continue
        LOG.info("--> Keeping all rows %s", ugc)
        for row in gdf.to_dict("records"):
            yield row


def workflow(pgconn, dt, filename):
    """Go Main Go"""
    zipfn = f"{filename}.zip"
    shpfn = do_download(zipfn)
    # track domain
    source = zipfn[:2].replace("_", "")
    LOG.info("Processing, using '%s' as the database source", source)
    df = read_shapefile(shpfn)
    # Compute the area and then sort to order duplicated UGCs :/
    # Database stores as sq km
    df["area2163"] = df["geometry"].to_crs(2163).area / 1e6
    df = df.sort_values(by="area2163", ascending=False)
    # Compute the ugc column
    wfocol = "CWA"
    if source in ["mz", "oz", "hz"]:
        df["STATE"] = ""
        df["ugc"] = df["ID"]
        wfocol = "WFO"
    elif source == "c":
        geo_type = "C"
        df["ugc"] = df["STATE"] + geo_type + df["FIPS"].str.slice(-3)
        df["NAME"] = df["COUNTYNAME"]
    else:
        geo_type = "Z"
        df["ugc"] = df["STATE"] + geo_type + df["ZONE"]
    df = df.rename(columns={wfocol: "wfo"})
    # Check that UGCs are not all null
    if df["ugc"].isna().all():
        LOG.info("Abort as all ugcs are null")
        sys.exit()
    postgis = gpd.read_postgis(
        sql_helper(
            "SELECT *, ctid from ugcs "
            "where end_ts is null and source = :source"
        ),
        pgconn,
        params={"source": source},
        geom_col="geom",
        index_col=None,  # We have dups at the UGC level
    )
    postgis["covered"] = False
    LOG.info(
        "Loaded %s '%s' type rows from the database",
        len(postgis.index),
        source,
    )
    # Rectify WFO
    df["wfo"] = df["wfo"].apply(lambda x: WFO_XREF.get(x, x))

    counts = {"new": 0, "changed": 0, "has": 0}
    # We need to dedup, but not in all cases, so alas
    for row in chunk_df(df):
        ugc = row["ugc"]
        name = row["NAME"].strip().replace("&", " and ")
        newugc = ugc not in postgis["ugc"].values
        current = postgis[
            (postgis["ugc"] == ugc)
            & (postgis["wfo"] == row["wfo"])
            & (postgis["name"] == name)
        ]
        # This should never happen, but just in case.
        if len(current.index) > 1:
            LOG.warning("FATAL ERROR, found multiple rows for %s", ugc)
            print(current)
            sys.exit()
        if len(current.index) == 1:
            # Do a crude area check
            if abs(row["area2163"] - current.iloc[0]["area2163"]) < 0.2:
                postgis.at[current.index[0], "covered"] = True
                counts["has"] += 1
                continue
            LOG.info(
                "%s updating sz diff %.2d -> %.2d",
                ugc,
                current.iloc[0]["area2163"],
                row["area2163"],
            )
            counts["changed"] += 1
        else:
            counts["new"] += 1

        LOG.info(
            "%s creating entry for %s[wfo=%s,isnew=%s]",
            ugc,
            name,
            row["wfo"],
            newugc,
        )

        # Finally, insert the new geometry
        pgconn.execute(
            sql_helper(
                "INSERT into ugcs (ugc, name, state, begin_ts, wfo, geom, "
                "source) VALUES (:ugc, :name, :state, :begints, :wfo, "
                "ST_Multi(ST_SetSRID(ST_GeomFromEWKT(:geom),4326)), :source)"
            ),
            {
                "ugc": ugc,
                "name": name,
                "state": row["STATE"],
                "begints": "1980-01-01" if newugc else dt,
                "wfo": row["wfo"],
                "geom": new_poly(row["geometry"]).wkt,
                "source": source,
            },
        )
    for _idx, row in postgis[~postgis["covered"]].iterrows():
        LOG.info("%s not found in update, truncating.", row["ugc"])
        truncate(pgconn, dt, row["ugc"], source, row["ctid"])

    LOG.info(
        "NEW: %s Updated: %s Has: %s",
        counts["new"],
        counts["changed"],
        counts["has"],
    )

    db_fixes(pgconn, dt)


@click.command()
@click.option(
    "--date",
    "dt",
    required=True,
    type=click.DateTime(),
    help="Date of the data",
)
@click.option("--filename", required=True, help="Zip file name (no extension)")
@click.option("--dryrun", is_flag=True, help="Dry run")
def main(dt: datetime, filename: str, dryrun: bool) -> None:
    """Go Main Go"""
    # Assumption is 18 UTC implementation timestamp
    dt = dt.replace(hour=18, tzinfo=timezone.utc)
    with get_sqlalchemy_conn("postgis") as pgconn:
        pgconn.row_factory = dict_row
        workflow(pgconn, dt, filename)
        if not dryrun:
            pgconn.commit()
        else:
            LOG.warning("---- DRY RUN, no commit ----")
    LOG.info("Done!")


if __name__ == "__main__":
    main()



"""Test alaska_marine workflow."""

import pytest

from pywwa.testing import get_example_file
from pywwa.workflows import alaska_marine


@pytest.mark.parametrize("database", ["postgis"])
def test_workflow(cursor):
    """Test failure in the wild."""
    alaska_marine.process_data(cursor, get_example_file("CWFAER_0.txt"))
    alaska_marine.process_data(cursor, get_example_file("CWFAER_1.txt"))
    alaska_marine.process_data(cursor, get_example_file("CWFAER_2.txt"))

    cursor.execute("select count(*) from warnings_2025 where wfo = 'AFC'")
    assert cursor.fetchone()["count"] == 24  # not really cross-checked



"""Reuse code from metar_parser."""

# 3rd party
import click
from pyiem.util import utc
from twisted.internet import reactor
from twisted.internet.endpoints import TCP4ServerEndpoint
from twisted.internet.protocol import Factory
from twisted.protocols.basic import LineReceiver

from pywwa import LOG, common
from pywwa.database import get_database, load_metar_stations
from pywwa.workflows import metar

IEMDB = get_database("iem")
ASOSDB = get_database("asos")


# pylint: disable=abstract-method
class AWOSProtocol(LineReceiver):
    """Our line receiver."""

    def lineReceived(self, line: bytes):
        """Do what we need to do."""
        line = line.decode("ascii", "ignore").replace("METAR ", "")
        reactor.callLater(0, process_line, line)


class AWOSFactory(Factory):
    """Our protocol factory."""

    def buildProtocol(self, addr):
        """Build our protocol."""
        return AWOSProtocol()


def process_line(line):
    """Process a line of METAR data!"""
    # create a faked noaaport text product
    text = (
        f"000 \r\r\nSAUS43 KISU {utc():%d%H%M}\r\r\nMETAR \r\r\n{line}\r\r\n"
    )
    metar.real_processor(text)


def ready(_):
    """Do what we need to do."""
    endpoint = TCP4ServerEndpoint(reactor, 4000)
    endpoint.listen(AWOSFactory())
    metar.cleandb()


@click.command(help=__doc__)
@common.init
@common.disable_xmpp
def main(*args, **kwargs):
    """Run once at startup"""
    df = IEMDB.runInteraction(load_metar_stations, metar.NWSLI_PROVIDER)
    df.addCallback(ready)
    df.addErrback(LOG.error)
    reactor.run()


if __name__ == "__main__":
    main()



"""Do the rotation action that some products need."""

import gzip
import os
import sys
from io import BytesIO

BASE = "/mesonet/ldmdata/"


def handler(fnbase: str, fmt: str, bio: BytesIO):
    """Do things."""
    if fmt == "tif.Z":
        for i in range(10, -1, -1):
            oldfp = f"{BASE}/{fnbase}{i}.{fmt}"
            newfp = f"{BASE}/{fnbase}{i + 1}.{fmt}"
            if os.path.isfile(oldfp):
                os.rename(oldfp, newfp)

        # Write out the compressed version verbatim
        with open(f"{BASE}/{fnbase}0.{fmt}", "wb") as fh:
            fh.write(bio.getvalue())

        # Create the uncompressed version
        bio.seek(0)
        with gzip.open(bio, "rb") as fh:
            data = fh.read()
        fmt = "tif"
    else:
        data = bio.getvalue()

    for i in range(10, -1, -1):
        oldfp = f"{BASE}/{fnbase}{i}.{fmt}"
        newfp = f"{BASE}/{fnbase}{i + 1}.{fmt}"
        if os.path.isfile(oldfp):
            os.rename(oldfp, newfp)

    with open(f"{BASE}/{fnbase}0.{fmt}", "wb") as fh:
        fh.write(data)


def main(argv):
    """Do SOmething"""
    fnbase = argv[1]
    fmt = argv[2]
    dirname = f"{BASE}/{os.path.dirname(fnbase)}"
    if not os.path.isdir(dirname):
        os.makedirs(dirname)
    with BytesIO(sys.stdin.buffer.read()) as bio:
        handler(fnbase, fmt, bio)


if __name__ == "__main__":
    main(sys.argv)



"""Split RR7 products into their individual stations, so for AFOS discovery.

It is unclear why this is being done, but will keep it going for now.  These
products are culled after 10 days per AFOS database cleaning and that this is
SHEF data.
"""

# stdlib
import re

import click
from pyiem.nws.product import TextProduct

from pywwa import common
from pywwa.database import get_database
from pywwa.ldm import bridge


def real_process(txn, data):
    """Process the data"""
    prod = TextProduct(
        data, parse_segments=False, utcnow=common.utcnow(), ugc_provider={}
    )

    data = data.replace("\r\r\n", "z")

    tokens = re.findall(r"(\.A [A-Z0-9]{3} .*?=)", data)

    for token in tokens:
        sql = (
            "INSERT into products (pil, data, entered, source, wmo, bbb) "
            "values(%s, %s, %s, %s, %s, %s)"
        )
        sqlargs = (
            f"RR7{token[3:6]}",
            token.replace("z", "\n"),
            prod.valid,
            prod.source,
            prod.wmo,
            prod.bbb,
        )
        txn.execute(sql, sqlargs)


@click.command(help=__doc__)
@common.init
@common.disable_xmpp
def main(*args, **kwargs):
    """Go"""
    bridge(real_process, dbpool=get_database("afos"))
